
# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from os import environ

from neuro_san.internals.graph.registry.agent_network import AgentNetwork


class MissingAgentCheck:
    """
    Convience and consolidation for checks against missing/misnamed agents.
    """

    @staticmethod
    def check_agent_network(agent_network: AgentNetwork, agent_name: str) -> AgentNetwork:
        """
        :param agent_network: The AgentNetwork to check
        :param agent_name: The name of the agent to use for the session.
        :return: The AgentNetwork corresponding to the agent_name
        """

        if agent_network is None:
            message = f"""
Agent named "{agent_name}" not found in manifest file: {environ.get("AGENT_MANIFEST_FILE")}.

Some things to check:
1. If the manifest file named above is None, know that the default points
   to the one provided with the neuro-san library for a smoother out-of-box
   experience.  If the agent you wanted is not part of that standard distribution,
   you need to set the AGENT_MANIFEST_FILE environment variable to point to a
   manifest.hocon file associated with your own project(s).
2. Check that the environment variable AGENT_MANIFEST_FILE is pointing to
   the manifest.hocon file that you expect and has no typos.
3. Does your manifest.hocon file contain a key for the agent specified?
4. Does the value for the key in the manifest file have a value of 'true'?
5. Does your agent name have a typo either in the hocon file or on the command line?
"""
            raise ValueError(message)

        return agent_network




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional

from copy import copy

from neuro_san.client.thinking_file_message_processor import ThinkingFileMessageProcessor
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.internals.messages.origination import Origination
from neuro_san.message_processing.basic_message_processor import BasicMessageProcessor
from neuro_san.session.direct_agent_session import DirectAgentSession


class StreamingInputProcessor:
    """
    Processes AgentCli input by using the neuro-san streaming API.
    """

    def __init__(self,
                 default_input: str = "",
                 thinking_file: str = None,
                 session: AgentSession = None,
                 thinking_dir: str = None):
        """
        Constructor
        """
        super().__init__()
        self.default_input: str = default_input
        self.session: AgentSession = session
        self.processor = BasicMessageProcessor()
        if thinking_dir is not None and thinking_file is not None:
            self.processor.add_processor(ThinkingFileMessageProcessor(thinking_file, thinking_dir))

        if self.session is None:
            raise ValueError("StreamingInputProcessor session cannot be None")

    def get_message_processor(self) -> BasicMessageProcessor:
        """
        :return: The message processor
        """
        return self.processor

    # pylint: disable=too-many-locals
    def process_once(self, state: Dict[str, Any]) -> Dict[str, Any]:
        """
        Use polling strategy to communicate with agent.
        :param state: The state dictionary to pass around
        :return: An updated state dictionary
        """
        empty: Dict[str, Any] = {}
        user_input: str = state.get("user_input")
        last_chat_response = state.get("last_chat_response")
        num_input = state.get("num_input")
        chat_context = state.get("chat_context", empty)
        chat_filter = state.get("chat_filter", empty)
        origin_str: str = ""

        if user_input is None or user_input == self.default_input:
            return state

        sly_data: Optional[Dict[str, Any]] = state.get("sly_data", None)
        # Note that by design, a client does not have to interpret the
        # chat_context at all. It merely needs to pass it along to continue
        # the conversation.
        chat_request: Dict[str, Any] = self.formulate_chat_request(user_input, sly_data,
                                                                   chat_context, chat_filter)
        self.reset()

        return_state: Dict[str, Any] = copy(state)
        returned_sly_data: Optional[Dict[str, Any]] = None
        chat_responses: Generator[Dict[str, Any], None, None] = self.session.streaming_chat(chat_request)
        for chat_response in chat_responses:

            response: Dict[str, Any] = chat_response.get("response", empty)

            self.processor.process_message(response)

            # Update the state if there is something to update it with
            chat_context = self.processor.get_chat_context()
            last_chat_response = self.processor.get_compiled_answer()
            returned_sly_data: Dict[str, Any] = self.processor.get_sly_data()
            origin_str = Origination.get_full_name_from_origin(self.processor.get_answer_origin())

        # Update the sly_data if new sly_data was returned
        if returned_sly_data is not None:
            if sly_data is not None:
                sly_data.update(returned_sly_data)
            else:
                sly_data = returned_sly_data.copy()

        if origin_str is None or len(origin_str) == 0:
            origin_str = "agent network"

        update = {
            "chat_context": chat_context,
            "num_input": num_input + 1,
            "last_chat_response": last_chat_response,
            "user_input": None,
            "sly_data": sly_data,
            "origin_str": origin_str,
            "token_accounting": self.processor.get_token_accounting()
        }
        return_state.update(update)

        return return_state

    def formulate_chat_request(self, user_input: str,
                               sly_data: Dict[str, Any] = None,
                               chat_context: Dict[str, Any] = None,
                               chat_filter: Dict[str, Any] = None) -> Dict[str, Any]:
        """
        Formulates a single chat request given the user_input
        :param user_input: The string to send
        :param sly_data: The sly_data dictionary to send
        :param chat_context: The chat context dictionary that allows the context of a
                    continuing conversation to be reconstructed on another server.
        :param chat_filter: The ChatFilter to apply to the request.
        :return: A dictionary representing the chat request to send
        """
        chat_request = {
            "user_message": {
                "type": ChatMessageType.HUMAN,
                "text": user_input
            }
        }

        if bool(chat_context):
            # Recall that non-empty dictionaries evaluate to True
            chat_request["chat_context"] = chat_context

        if sly_data is not None and len(sly_data.keys()) > 0:
            chat_request["sly_data"] = sly_data

        if chat_filter is not None and len(chat_filter.keys()) > 0:
            chat_request["chat_filter"] = chat_filter

        return chat_request

    def reset(self):
        """
        Reset for a new exchange
        """
        # Reset the message processor to receive a new answer
        self.processor.reset()

        if isinstance(self.session, DirectAgentSession):
            # Direct sessions need their Origination reset otherwise chat_context
            # origins do not match up.
            self.session.reset()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor

from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.origination import Origination


class InvocationContext:
    """
    Interface for encapsulating specific policy classes that pertain to
    a single invocation of an AgentSession or AsyncAgentSession, whether by way of a
    service call or library call.
    """

    def start(self):
        """
        Starts the active components of this invocation context.
        Do this separately from constructor for more control.
        """
        raise NotImplementedError

    def get_async_session_factory(self) -> AsyncAgentSessionFactory:
        """
        :return: The AsyncAgentSessionFactory associated with the invocation
        """
        raise NotImplementedError

    def get_asyncio_executor(self) -> AsyncioExecutor:
        """
        :return: The AsyncioExecutor associated with the invocation
        """
        raise NotImplementedError

    def get_origination(self) -> Origination:
        """
        :return: The Origination instance carrying state about tool instantation
                during the course of the AgentSession.
        """
        raise NotImplementedError

    def get_journal(self) -> Journal:
        """
        :return: The Journal instance that allows message reporting
                during the course of the AgentSession.
        """
        raise NotImplementedError

    def get_queue(self) -> AsyncCollatingQueue:
        """
        :return: The AsyncCollatingQueue instance via which messages are streamed to the
                AgentSession mechanics
        """
        raise NotImplementedError

    def get_metadata(self) -> Dict[str, str]:
        """
        :return: The metadata to pass along with any request
        """
        raise NotImplementedError

    def close(self):
        """
        Release resources owned by this context
        """
        raise NotImplementedError

    def get_request_reporting(self) -> Dict[str, Any]:
        """
        :return: The request reporting dictionary
        """
        raise NotImplementedError

    def get_llm_factory(self) -> ContextTypeLlmFactory:
        """
        :return: The ContextTypeLlmFactory instance for the session
        """
        raise NotImplementedError

    def get_toolbox_factory(self) -> ContextTypeToolboxFactory:
        """
        :return: The ContextTypeToolboxFactory instance for the session
        """
        raise NotImplementedError

    def get_reservationist(self) -> Reservationist:
        """
        :return: The Reservationist instance for the session
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Dict
from typing import List

import logging

from leaf_server_common.server.server_lifetime import ServerLifetime
from leaf_server_common.server.server_loop_callbacks import ServerLoopCallbacks

from neuro_san.api.grpc import agent_pb2
from neuro_san.api.grpc import concierge_pb2_grpc

from neuro_san.internals.interfaces.agent_state_listener import AgentStateListener
from neuro_san.internals.interfaces.agent_storage_source import AgentStorageSource
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.service.generic.agent_server_logging import AgentServerLogging
from neuro_san.service.grpc.agent_servicer_to_server import AgentServicerToServer
from neuro_san.service.grpc.concierge_service import ConciergeService
from neuro_san.service.grpc.dynamic_agent_router import DynamicAgentRouter
from neuro_san.service.grpc.grpc_agent_service import GrpcAgentService
from neuro_san.service.interfaces.agent_server import AgentServer
from neuro_san.service.utils.server_context import ServerContext
from neuro_san.service.utils.server_status import ServerStatus
from neuro_san.session.agent_service_stub import AgentServiceStub

DEFAULT_SERVER_NAME: str = 'neuro-san.Agent'
DEFAULT_SERVER_NAME_FOR_LOGS: str = 'Agent Server'
DEFAULT_MAX_CONCURRENT_REQUESTS: int = 10

# Better that we kill ourselves than kubernetes doing it for us
# in the middle of a request if there are resource leaks.
# This is per the lifetime of the server (before it kills itself).
DEFAULT_REQUEST_LIMIT: int = 1000 * 1000


# pylint: disable=too-many-instance-attributes
class GrpcAgentServer(AgentServer, AgentStateListener):
    """
    Server implementation for the Agent gRPC Service.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self, port: int,
                 server_loop_callbacks: ServerLoopCallbacks,
                 server_context: ServerContext,
                 server_name: str = DEFAULT_SERVER_NAME,
                 server_name_for_logs: str = DEFAULT_SERVER_NAME_FOR_LOGS,
                 max_concurrent_requests: int = DEFAULT_MAX_CONCURRENT_REQUESTS,
                 request_limit: int = DEFAULT_REQUEST_LIMIT,
                 forwarded_request_metadata: str = AgentServer.DEFAULT_FORWARDED_REQUEST_METADATA):
        """
        Constructor

        :param port: The integer port number for the service to listen on
        :param server_loop_callbacks: The ServerLoopCallbacks instance for
                break out methods in main serving loop.
        :param server_context: ServerContext carrying global-ish state
        :param server_name: The name of the service
        :param server_name_for_logs: The name of the service for log files
        :param max_concurrent_requests: The maximum number of requests to handle at a time.
        :param request_limit: The number of requests to service before shutting down.
                        This is useful to be sure production environments can handle
                        a service occasionally going down.
        :param forwarded_request_metadata: A space-delimited list of http metadata request keys
                        to forward to logs/other requests
        """
        self.port = port
        self.server_loop_callbacks = server_loop_callbacks

        self.server_logging = AgentServerLogging(server_name_for_logs, forwarded_request_metadata)
        self.server_logging.setup_logging()

        self.logger = logging.getLogger(__name__)

        # Below is now odd.
        self.server_name: str = server_name
        self.server_name_for_logs: str = server_name_for_logs
        self.max_concurrent_requests: int = max_concurrent_requests
        self.request_limit: int = request_limit
        self.server_context: ServerContext = server_context

        self.server_lifetime = None
        self.security_cfg = None
        self.services: List[GrpcAgentService] = []
        self.service_router: DynamicAgentRouter = DynamicAgentRouter()

    def get_services(self) -> List[GrpcAgentService]:
        """
        :return: A list of the AgentServices being served up by this instance
        """
        return self.services

    def agent_added(self, agent_name: str, source: AgentStorageSource):
        """
        Agent is being added to the service.
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        service = GrpcAgentService(self.server_lifetime,
                                   self.security_cfg,
                                   agent_name,
                                   source.get_agent_network_provider(agent_name),
                                   self.server_logging,
                                   self.server_context)
        self.services.append(service)
        servicer_to_server = AgentServicerToServer(service)
        agent_rpc_handlers = servicer_to_server.build_rpc_handlers()
        agent_service_name: str = AgentServiceStub.prepare_service_name(agent_name)
        self.service_router.add_service(agent_service_name, agent_rpc_handlers)

    def agent_modified(self, agent_name: str, source: AgentStorageSource):
        """
        Agent is being modified in the service scope.
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        # Endpoints configuration has not changed,
        # so nothing to do here, actually.
        _ = agent_name

    def agent_removed(self, agent_name: str, source: AgentStorageSource):
        """
        Agent is being removed from the service.
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        agent_service_name: str = AgentServiceStub.prepare_service_name(agent_name)
        self.service_router.remove_service(agent_service_name)

    def prepare_for_serving(self):
        """
        Prepare server for running.
        """
        values = agent_pb2.DESCRIPTOR.services_by_name.values()
        self.server_lifetime = ServerLifetime(
            self.server_name,
            self.server_name_for_logs,
            self.port, self.logger,
            request_limit=self.request_limit,
            max_workers=self.max_concurrent_requests,
            max_concurrent_rpcs=None,
            # Used for health checking. Probably needs agent-specific love.
            protocol_services_by_name_values=values,
            loop_sleep_seconds=5.0,
            server_loop_callbacks=self.server_loop_callbacks)

        server = self.server_lifetime.create_server()

        # New-style service
        self.security_cfg = None     # ... yet

        # Add listener to handle adding per-agent gRPC services
        # to our dynamic router:
        network_storage_dict: Dict[str, AgentNetworkStorage] = self.server_context.get_network_storage_dict()
        public_storage: AgentNetworkStorage = network_storage_dict.get("public")
        for network_storage in network_storage_dict.values():
            network_storage.add_listener(self)

        # Add DynamicAgentRouter instance as a generic RPC handler for our server:
        server.add_generic_rpc_handlers((self.service_router,))

        concierge_service: ConciergeService = \
            ConciergeService(self.server_lifetime,
                             self.security_cfg,
                             self.server_logging,
                             public_storage)
        concierge_pb2_grpc.add_ConciergeServiceServicer_to_server(
            concierge_service,
            server)

    def serve(self):
        """
        Start serving gRPC requests
        """
        server_status: ServerStatus = self.server_context.get_server_status()
        server_status.grpc_service.set_status(True)
        self.server_lifetime.run()

    def stop(self):
        """
        Stop the server.
        """
        server_status: ServerStatus = self.server_context.get_server_status()
        server_status.grpc_service.set_status(False)
        # pylint: disable=protected-access
        self.server_lifetime._stop_serving()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from leaf_common.config.config_filter_chain import ConfigFilterChain

from neuro_san.internals.graph.persistence.manifest_dict_config_filter import ManifestDictConfigFilter
from neuro_san.internals.graph.persistence.manifest_key_config_filter import ManifestKeyConfigFilter
from neuro_san.internals.graph.persistence.served_manifest_config_filter import ServedManifestConfigFilter


class ManifestFilterChain(ConfigFilterChain):
    """
    ConfigFilterChain for manifest files
    """

    def __init__(self, manifest_file: str):
        """
        Constructor

        :param manifest_file: The name of the manifest file we are processing for logging purposes
        """
        super().__init__()

        # Order matters
        self.register(ManifestKeyConfigFilter(manifest_file))
        self.register(ManifestDictConfigFilter(manifest_file))
        self.register(ServedManifestConfigFilter(manifest_file))




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from urllib.parse import ParseResult
from urllib.parse import urlparse


class ExternalAgentParsing:
    """
    Class handles parsing references to an external agent server
    so that its agents can be used as tools.
    """

    @staticmethod
    def parse_external_agent(agent_url: str) -> Dict[str, str]:
        """
        :param agent_url: The URL describing where to find the desired agent.
        :return: A Dictionary with the following keys:
                "host" - the hostname where the agent lives
                "port" - the port on the host which serves up the agent (if any)
                "agent_name" - the name of the agent on that host

                OR

                None if the parsing of the agent_url was unsuccessful.
        """
        if agent_url is None or len(agent_url) == 0:
            return None

        parse_result: ParseResult = urlparse(agent_url)
        if parse_result is None:
            return None

        if parse_result.path is None or len(parse_result.path) <= 1:
            # We don't have enough characters in the path to even specify
            # an agent that lives on the same server.
            return None

        if not parse_result.path.startswith("/"):
            # This is not an external agent specification
            return None

        host: str = None
        port: str = None
        if len(parse_result.netloc) > 0:
            # We have a host specified
            split: List[str] = parse_result.netloc.split(":")
            host = split[0]
            if len(split) > 1:
                port = split[1]

        # Special case for detecting localhost
        if host is None or len(host) == 0:
            host = "localhost"

        # Get the agent name from the URL by looking at the path
        # Remove any leading slashes from the path for the agent name.
        # Note: While we need to get the agent name for proper gRPC routing,
        #       this is not yet super robust against any non-default case
        #       where some other entity needs a non-standard path for routing
        #       (like a load balancer).  Cross that bridge when we get to it.
        agent_name: str = parse_result.path
        while agent_name.startswith("/"):
            agent_name = agent_name[1:]

        # Assemble the return dictionary
        return_dict = {
            "host": host,
            "port": port,
            "agent_name": agent_name,
        }
        return return_dict

    @staticmethod
    def is_external_agent(agent_url: str) -> bool:
        """
        :param agent_url: The URL describing where to find the desired agent.
        :return: True if the given string is interpretable as an agent url
                (without actually connecting to it).  False otherwise.
        """
        agent_location: Dict[str, str] = ExternalAgentParsing.parse_external_agent(agent_url)
        is_external: bool = agent_location is not None
        return is_external

    @staticmethod
    def is_mcp_tool(tool_ref: Union[str, Dict[str, Any]]) -> bool:
        """
        Check if the tool reference is for an MCP server.
        :param tool_ref: String URL or dict config for MCP tool
        :return: True if this is an MCP tool reference
        """
        # Support both str and dict format;
        # str format: "https://mcp.deepwiki.com/mcp" or "http://localhost:8000/mcp/"
        # dict format:
        # {
        #       "url": "https://mcp.deepwiki.com/mcp",
        #       "tools": ["read_wiki_structure", "ask_question"],
        # }

        # If it is a dict, it is assumed it is MCP for now.
        # This may change in the future when Neuro-SAN supports other protocals like A2A.
        if isinstance(tool_ref, dict):
            return True

        if isinstance(tool_ref, str):
            return (tool_ref.startswith("https://mcp") or tool_ref.endswith(("/mcp", "/mcp/")))

        return False

    @staticmethod
    def get_safe_agent_name(agent_url: str) -> str:
        """
        :param agent_url: The URL describing where to find the desired agent.
        :return: A name that is suitable for using within agent toolkits (like langchain)
                for internal tool reference.
        """
        safe_name: str = agent_url
        if ExternalAgentParsing.is_external_agent(agent_url):

            agent_location: Dict[str, str] = ExternalAgentParsing.parse_external_agent(agent_url)

            # FWIW: langchain internal tool references must satisfy the regex: "^[a-zA-Z0-9_-]+$"
            # It's possible that more complex external references might have the agent_name
            # needing further mangling.  Cross that bridge when we have a real example.
            # As a part of valid URL, agent_name can only have "/" in it.
            safe_name = "__" + agent_location.get("agent_name").replace("/", "__")

        return safe_name




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import copy
import threading

from langchain_core.tools import BaseTool
from langchain_mcp_adapters.client import MultiServerMCPClient
from neuro_san.internals.run_context.langchain.mcp.mcp_clients_info_restorer import McpClientsInfoRestorer


class LangChainMcpAdapter:
    """
    Adapter class to fetch tools from a Multi-Client Protocol (MCP) server and return them as
    LangChain-compatible tools. This class provides static methods for interacting with MCP servers.
    """

    _mcp_info_lock: threading.Lock = threading.Lock()
    _mcp_clients_info: Dict[str, Any] = None

    def __init__(self):
        """
        Constructor
        """
        self.client_allowed_tools: List[str] = []

    @staticmethod
    def _load_mcp_clients_info():
        """
        Loads MCP clients information from a configuration file if not already loaded.
        """
        with LangChainMcpAdapter._mcp_info_lock:
            if LangChainMcpAdapter._mcp_clients_info is None:
                LangChainMcpAdapter._mcp_clients_info = McpClientsInfoRestorer().restore()
                if LangChainMcpAdapter._mcp_clients_info is None:
                    # Something went wrong reading the file.
                    # Prevent further attempts to load info.
                    LangChainMcpAdapter._mcp_clients_info = {}

    async def get_mcp_tools(
            self,
            server_url: str,
            allowed_tools: Optional[List[str]] = None,
    ) -> List[BaseTool]:
        """
        Fetches tools from the given MCP server and returns them as a list of LangChain-compatible tools.

        :param server_url: URL of the MCP server, e.g. https://mcp.deepwiki.com/mcp or http://localhost:8000/mcp/
        :param allowed_tools: Optional list of tool names to filter from the server's available tools.
                              If None, all tools from the server will be returned.

        :return: A list of LangChain BaseTool instances retrieved from the MCP server.
        """
        if self._mcp_clients_info is None:
            self._load_mcp_clients_info()

        mcp_tool_dict: Dict[str, Any] = {
            "url": server_url,
            "transport": "streamable_http",
        }
        # Try to look up authentication details from the URL
        headers_dict: Dict[str, Any] =\
            self._mcp_clients_info.get(server_url, {}).get("headers")
        if headers_dict:
            mcp_tool_dict["headers"] = copy.copy(headers_dict)

        client = MultiServerMCPClient(
            {"server": mcp_tool_dict}
        )

        # The get_tools() method returns a list of StructuredTool instances, which are subclasses of BaseTool.
        # Internally, it calls load_mcp_tools(), which uses an `async with create_session(...)` block.
        # This guarantees that any temporary MCP session created is properly closed when the block exits,
        # even if an error is raised during tool loading.
        # See: https://github.com/langchain-ai/langchain-mcp-adapters/blob/main/langchain_mcp_adapters/tools.py#L164
        # Optimization:
        #   It's possible we might want to cache these results somehow to minimize tool calls.
        mcp_tools: List[BaseTool] = await client.get_tools()

        # If allowed_tools is provided, filter the list to include only those tools.
        client_allowed_tools: List[str] = allowed_tools
        if client_allowed_tools is None:
            # Check if MCP client info has a "tools" field to use as allowed tools.
            client_allowed_tools = self._mcp_clients_info.get(server_url, {}).get("tools", [])
        # If client allowed tools is an empty list, do not filter the tools.
        if client_allowed_tools:
            mcp_tools = [tool for tool in mcp_tools if tool.name in client_allowed_tools]

        self.client_allowed_tools = client_allowed_tools

        for tool in mcp_tools:
            # Add "langchain_tool" tags so journal callback can idenitify it.
            # These MCP tools are treated as Langchain tools and can be reported in the thinking file.
            tool.tags = ["langchain_tool"]

        return mcp_tools




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.graph.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.graph.interfaces.callable_activation import CallableActivation
from neuro_san.internals.run_context.interfaces.agent_network_inspector import AgentNetworkInspector
from neuro_san.internals.run_context.interfaces.run_context import RunContext


class AbstractCallableActivation(CallableActivation):
    """
    An abstract implementation of the CallableActivation interface
    containing common policy for all tools.

    Worth noting that this is used as a base implementation for:
        * ClassActivation
        * ExternalActivation
        * CallingActivation
    """

    def __init__(self,
                 factory: AgentToolFactory,
                 agent_tool_spec: Dict[str, Any],
                 sly_data: Dict[str, Any]):
        """
        Constructor

        :param factory: The factory for Agent Tools.
        :param agent_tool_spec: The dictionary describing the JSON agent tool
                            to be used by the instance
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        """
        self.factory: AgentToolFactory = factory
        self.agent_tool_spec: Dict[str, Any] = agent_tool_spec
        self.sly_data: Dict[str, Any] = sly_data

        # Subclasses should set up the RunContext for themselves and get the journal from it
        # because not everyone needs an llm_config
        self.run_context: RunContext = None

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        return self.agent_tool_spec

    def get_name(self) -> str:
        """
        :return: the name of the data-driven agent as it comes from the spec
        """
        agent_spec: Dict[str, Any] = self.get_agent_tool_spec()
        agent_name: str = self.factory.get_name_from_spec(agent_spec)
        return agent_name

    def get_inspector(self) -> AgentNetworkInspector:
        """
        :return: The factory containing all the tool specs
        """
        # For now, our inspector is an AgentToolFactory
        return self.factory

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        return self.run_context.get_origin()

    async def delete_resources(self, parent_run_context: RunContext):
        """
        Cleans up after any allocated resources on their server side.
        :param parent_run_context: The RunContext which contains the scope
                    of operation of this CallableActivation
        """
        if self.run_context is not None:
            await self.run_context.delete_resources(parent_run_context)
            self.run_context = None

    async def build(self) -> BaseMessage:
        """
        Main entry point to the class.

        :return: A BaseMessage produced during this process.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""

DEFAULT_HTTP_CONNECTIONS_BACKLOG: int = 128
DEFAULT_HTTP_IDLE_CONNECTIONS_TIMEOUT_SECONDS: int = 3600
DEFAULT_HTTP_SERVER_INSTANCES: int = 1
DEFAULT_HTTP_SERVER_MONITOR_INTERVAL_SECONDS: int = 0


class HttpServerConfig:
    """
    Class aggregating Tornado http server run-time configuration parameters.
    """

    def __init__(self):
        self.http_connections_backlog: int = DEFAULT_HTTP_CONNECTIONS_BACKLOG
        self.http_idle_connection_timeout_seconds: int = DEFAULT_HTTP_IDLE_CONNECTIONS_TIMEOUT_SECONDS
        self.http_server_instances: int = DEFAULT_HTTP_SERVER_INSTANCES
        self.http_port: int = 80
        self.http_server_monitor_interval_seconds: int = DEFAULT_HTTP_SERVER_MONITOR_INTERVAL_SECONDS




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT


class AgentServer:
    """
    Interface for an AgentServer, regardless of transport mechanism
    """

    # A space-delimited list of http metadata request keys to forward to logs/other requests
    DEFAULT_FORWARDED_REQUEST_METADATA: str = "request_id user_id"

    def stop(self):
        """
        Stop the server.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

import json
import logging

from json.decoder import JSONDecodeError
from pyparsing.exceptions import ParseException
from pyparsing.exceptions import ParseSyntaxException

from leaf_common.persistence.easy.easy_hocon_persistence import EasyHoconPersistence
from leaf_common.persistence.interface.restorer import Restorer


class RawManifestRestorer(Restorer):
    """
    Implementation of the Restorer interface that reads the contents
    of a single manifest file for agent networks/registries.
    """

    def __init__(self):
        """
        Constructor
        """
        self.logger = logging.getLogger(self.__class__.__name__)

    def restore(self, file_reference: str = None) -> Dict[str, Any]:
        """
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: a built map of agent networks
        """

        one_manifest: Dict[str, Any] = {}

        if file_reference.endswith(".hocon"):
            hocon = EasyHoconPersistence()
            try:
                one_manifest = hocon.restore(file_reference=file_reference)
            except (ParseException, ParseSyntaxException) as exception:
                message: str = f"""
There was an error parsing the agent network manifest file "{file_reference}".
See the accompanying ParseException (above) for clues as to what might be
syntactically incorrect in that file.
"""
                raise ParseException(message) from exception
        else:
            try:
                with open(file_reference, "r", encoding="utf-8") as json_file:
                    one_manifest = json.load(json_file)
            except FileNotFoundError:
                # Use the common verbiage below
                one_manifest = None
            except JSONDecodeError as exception:
                message: str = f"""
There was an error parsing the agent network manifest file "{file_reference}".
See the accompanying JSONDecodeError exception (above) for clues as to what might be
syntactically incorrect in that file.
"""
                raise ParseException(message) from exception

        if one_manifest is None:
            message = f"Could not find manifest file at path: {file_reference}.\n" + """
Some common problems include:
* The file itself simply does not exist.
* Path is not an absolute path and you are invoking the server from a place
  where the path is not reachable.
* The path has a typo in it.

Double-check the value of the AGENT_MANIFEST_FILE env var and
your current working directory (pwd).
"""
            raise FileNotFoundError(message)

        return one_manifest




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from leaf_common.config.file_of_class import FileOfClass

# Normally we don't use __init__.py files to define anything,
# but here we define some constants that point to important directories in the distribution.
TOP_LEVEL_DIR = FileOfClass(__file__)
DEPLOY_DIR = FileOfClass(__file__, path_to_basis="./deploy")
REGISTRIES_DIR = FileOfClass(__file__, path_to_basis="./registries")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from collections.abc import Sequence
from typing import Any
from typing import Dict
from typing import List

from pydantic import ConfigDict

from langchain_core.agents import AgentAction
from langchain_core.agents import AgentFinish
from langchain_core.callbacks.base import AsyncCallbackHandler
from langchain_core.documents import Document
from langchain_core.messages import ToolMessage
from langchain_core.messages.base import BaseMessage
from langchain_core.outputs import LLMResult
from langchain_core.outputs.chat_generation import ChatGeneration

from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.journals.originating_journal import OriginatingJournal
from neuro_san.internals.journals.tool_argument_reporting import ToolArgumentReporting
from neuro_san.internals.messages.agent_message import AgentMessage
from neuro_san.internals.messages.agent_tool_result_message import AgentToolResultMessage
from neuro_san.internals.messages.origination import Origination


# pylint: disable=too-many-ancestors
class JournalingCallbackHandler(AsyncCallbackHandler):
    """
    AsyncCallbackHandler implementation that intercepts agent-level chatter

    We use this guy to intercept agent-level messages like:
        "Thought: Do I need a tool?" and preliminary results from the agent

    We are currently only listening to on_llm_end(), but there are many other
    callbacks to hook into, most of which are not really productive.
    Some are overriden here to explore, others are not.  See the base
    AsyncCallbackHandler to explore more.

    Of note: This CallbackHandler mechanism is the kind of thing that
            LoggingCallbackHandler hooks into to produce egregiously chatty logs.
    """

    # Declarations of member variables here satisfy Pydantic style,
    # which is a type validator that langchain is based on which
    # is able to use JSON schema definitions to validate fields.
    journal: OriginatingJournal

    # This guy needs to be a pydantic class and in order to have
    # a non-pydantic Journal as a member, we need to do this.
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
            self,
            calling_agent_journal: Journal,
            base_journal: Journal,
            parent_origin: List[Dict[str, Any]],
            origination: Origination
    ):
        """
        Constructor

        :param calling_agent_journal: The journal of the calling agent
        :param base_journal: The Journal instance that allows message reporting during the course of the AgentSession.
            This is used to construct the langchain_tool_journal.
        :param parent_origin: A List of origin dictionaries indicating the origin of the run
            This is used to construct the langchain_tool_journal.
        :param origination: The Origination instance carrying state about tool instantation
            during the course of the AgentSession. This is used to construct the langchain_tool_journal.
        """

        # The calling-agent journal logs the execution flow from the perspective of the agent invoking the tool
        # (e.g., MusicNerdPro). In contrast, the LangChain tool journal represents the tool's own execution
        # context—similar to how coded tools like Accountant have their own journal tied to their run context.

        # LangChain tools don’t instantiate their own RunContext, so they lack a dedicated journal by default.
        # To maintain consistency with how other tools are tracked, we explicitly create a langchain_tool_journal
        # when the tool starts. This ensures tool-specific inputs and outputs are captured independently,
        # while still allowing the calling agent to log its own perspective.

        self.calling_agent_journal: Journal = calling_agent_journal
        self.base_journal: Journal = base_journal
        self.parent_origin: List[Dict[str, Any]] = parent_origin
        self.origination: Origination = origination
        self.langchain_tool_journal: Journal = None
        self.origin: List[Dict[str, Any]] = None

    async def on_llm_end(self, response: LLMResult,
                         **kwargs: Any) -> None:
        # Empirically we have seen that LLMResults that come in on_llm_end() calls
        # have a generations field which is a list of lists. Inside that inner list,
        # the first object is a ChatGeneration, whose text field tends to have agent
        # thinking in it.
        generations = response.generations[0]
        first_generation = generations[0]
        if isinstance(first_generation, ChatGeneration):
            content: str = first_generation.text
            if content is not None and len(content) > 0:
                # Package up the thinking content as an AgentMessage to stream
                message = AgentMessage(content=content.strip())
                # Some AGENT messages that come from this source end up being dupes
                # of AI messages that can come later.
                # Use this method to put the message on hold for later comparison.
                await self.calling_agent_journal.write_message_if_next_not_dupe(message)

    async def on_chain_end(self, outputs: Dict[str, Any],
                           **kwargs: Any) -> None:
        # print(f"In on_chain_end() with {outputs}")
        return

    async def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        tags: List[str] = None,
        inputs: Dict[str, Any] = None,
        **kwargs: Any
    ) -> None:
        """
        Callback triggered when a tool starts execution.

        If the tool is identified as a LangChain tool (via the "langchain_tool" tag),
        this method creates a journal entry containing the tool's input arguments,
        origin metadata, and full tool name.

        :param serialized: Serialized representation of the tool, including its name and description.
        :param input_str: String representation of the tool's input.
        :param tags: List of tags associated with the tool. Used to determine whether it is a LangChain tool.
        :param inputs: Structured dictionary of input arguments
            passed to the tool.
        """

        # Extract tool name from the serialized data
        agent_name: str = serialized.get("name")

        # Remove any policy objects from the arguments passed in.
        tool_start_dict: Dict[str, Any] = ToolArgumentReporting.prepare_tool_start_dict(inputs)
        caller_structure: Dict[str, Any] = {
            "invoking_start": True,
            "invoked_agent_name": agent_name,
            "params": tool_start_dict.get("tool_args")
        }

        # Report that we are about to invoke a tool.
        message: BaseMessage = AgentMessage(content=f"Invoking: `{agent_name}` with:",
                                            structure=caller_structure)
        await self.calling_agent_journal.write_message(message)

        if "langchain_tool" in tags:

            # Build the origin path
            self.origin: List[Dict[str, Any]] = self.origination.add_spec_name_to_origin(self.parent_origin, agent_name)

            # Re-build the tool start dictionary we will report with tool origin information
            # for the langchain tool's journal.
            tool_start_dict = ToolArgumentReporting.prepare_tool_start_dict(inputs, self.origin)

            # Create a journal entry for this invocation and log the combined inputs
            self.langchain_tool_journal = OriginatingJournal(self.base_journal, self.origin)
            message = AgentMessage(content="Received arguments:", structure=tool_start_dict)
            await self.langchain_tool_journal.write_message(message)

    async def on_tool_end(self, output: Any, tags: List[str] = None, **kwargs: Any) -> None:
        """
        Callback triggered when a tool finishes execution.

        If the tool is identified as a LangChain tool (via the "langchain_tool" tag),
        this method logs the tool's output to both the calling agent's journal and the
        LangChain tool's specific journal.

        :param output: The result produced by the tool after execution.
        :param tags: List of tags associated with the tool. Used to determine whether it is a LangChain tool.
        """

        if "langchain_tool" in tags:
            # Log the tool output to the calling agent's journal
            # Note that output is changed to str here since it can be anything including a ToolMessage
            # If that is the case, the content is taken as the output.
            if isinstance(output, ToolMessage):
                output = output.content

            await self.calling_agent_journal.write_message(
                AgentToolResultMessage(content=str(output), tool_result_origin=self.origin)
            )

            # Also log the tool output to the LangChain tool-specific journal
            output_dict: Dict[str, Any] = {
                "tool_end": True,
                "tool_output": output
            }
            message: BaseMessage = AgentMessage(content="Got result:", structure=output_dict)
            await self.langchain_tool_journal.write_message(message)

    async def on_agent_action(self, action: AgentAction,
                              **kwargs: Any) -> None:
        # print(f"In on_agent_action() with {action}")
        return

    async def on_agent_finish(self, finish: AgentFinish,
                              **kwargs: Any) -> None:
        # print(f"In on_agent_finish() with {finish}")
        return

    async def on_retriever_end(self, documents: Sequence[Document],
                               **kwargs: Any) -> None:
        # print(f"In on_retriever_end() with {documents}")
        return




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Union

from langchain_core.messages.base import BaseMessage


class AgentMessage(BaseMessage):
    """
    BaseMessage implementation of a message from an agent
    """

    type: Literal["agent"] = "agent"

    def __init__(self, content: Union[str, List[Union[str, Dict]]] = "",
                 structure: Dict[str, Any] = None, **kwargs: Any) -> None:
        """
        Pass in content as positional arg.

        Args:
            content: The string contents of the message.
            structure: A dictionary to pack into the message
            kwargs: Additional fields to pass to the
        """
        super().__init__(content=content, **kwargs)
        self.structure: Dict[str, Any] = structure




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict

from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.http.handlers.base_request_handler import BaseRequestHandler


class FunctionHandler(BaseRequestHandler):
    """
    Handler class for neuro-san "function" API call.
    """

    async def get(self, agent_name: str):
        """
        Implementation of GET request handler for "function" API call.
        """
        metadata: Dict[str, Any] = self.get_metadata()
        service: AsyncAgentService = await self.get_service(agent_name, metadata)
        if service is None:
            return

        self.application.start_client_request(metadata, f"{agent_name}/function")
        try:
            data: Dict[str, Any] = {}
            result_dict: Dict[str, Any] = await service.function(data, metadata)

            # Return service response to the HTTP client
            self.set_header("Content-Type", "application/json")
            self.write(result_dict)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.process_exception(exc)
        finally:
            self.do_finish()
            self.application.finish_client_request(metadata, f"{agent_name}/function")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from neuro_san.service.generic.async_agent_service_provider import AsyncAgentServiceProvider


class AgentAuthorizer:
    """
    Abstract interface implementing some policy
    of allowing to route incoming requests to an agent.
    """

    def allow(self, agent_name) -> AsyncAgentServiceProvider:
        """
        :param agent_name: name of an agent
        :return: instance of AsyncAgentService if routing requests is allowed for this agent;
                 None otherwise
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

import json

from pyparsing.exceptions import ParseException
from pyparsing.exceptions import ParseSyntaxException

from leaf_common.persistence.easy.easy_hocon_persistence import EasyHoconPersistence
from leaf_common.persistence.interface.restorer import Restorer

from neuro_san import TOP_LEVEL_DIR


class LlmInfoRestorer(Restorer):
    """
    Implementation of the Restorer interface to read in an LlmInfo dictionary
    instance given a hocon file name.
    """

    def restore(self, file_reference: str = None):
        """
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: an object from some persisted store
        """
        config: Dict[str, Any] = None

        use_file: str = file_reference

        if file_reference is None or len(file_reference) == 0:
            # Read from the default
            use_file = TOP_LEVEL_DIR.get_file_in_basis("internals/run_context/langchain/llms/default_llm_info.hocon")

        try:
            if use_file.endswith(".json"):
                config = json.load(use_file)
            elif use_file.endswith(".hocon"):
                hocon = EasyHoconPersistence(full_ref=use_file, must_exist=True)
                config = hocon.restore()
            else:
                raise ValueError(f"file_reference {use_file} must be a .json or .hocon file")
        except (ParseException, ParseSyntaxException, json.decoder.JSONDecodeError) as exception:
            message = f"""
There was an error parsing the llm_info file "{use_file}".
See the accompanying ParseException (above) for clues as to what might be
syntactically incorrect in that file.
"""
            raise ParseException(message) from exception

        return config




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Generator

import asyncio
import json

from aiohttp import ClientPayloadError
from aiohttp import ClientOSError
from aiohttp import ClientSession
from aiohttp import ClientTimeout

from neuro_san.interfaces.async_agent_session import AsyncAgentSession
from neuro_san.session.abstract_http_service_agent_session import AbstractHttpServiceAgentSession


class AsyncHttpServiceAgentSession(AbstractHttpServiceAgentSession, AsyncAgentSession):
    """
    Implementation of AsyncAgentSession that talks to an HTTP service.
    """

    async def function(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the FunctionRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the FunctionResponse
                    protobufs structure. Has the following keys:
                "function" - the dictionary description of the function
        """
        path: str = self.get_request_path("function")
        result_dict: Dict[str, Any] = None
        try:
            timeout: ClientTimeout = None
            if self.timeout_in_seconds is not None:
                timeout = ClientTimeout(self.timeout_in_seconds)

            async with ClientSession(headers=self.get_headers(),
                                     timeout=timeout
                                     ) as session:
                async with session.get(path, json=request_dict) as response:
                    result_dict = await response.json()
                    return result_dict
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc

    async def connectivity(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConnectivityRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConnectivityResponse
                    protobufs structure. Has the following keys:
                "connectivity_info" - the list of connectivity descriptions for
                                    each node in the agent network the service
                                    wants the client ot know about.
        """
        path: str = self.get_request_path("connectivity")
        result_dict: Dict[str, Any] = None
        try:
            timeout: ClientTimeout = None
            if self.timeout_in_seconds is not None:
                timeout = ClientTimeout(self.timeout_in_seconds)
            async with ClientSession(headers=self.get_headers(),
                                     timeout=timeout
                                     ) as session:
                async with session.get(path, json=request_dict) as response:
                    result_dict = await response.json()
                    return result_dict
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc

    async def streaming_chat(self, request_dict: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """
        :param request_dict: A dictionary version of the ChatRequest
                    protobufs structure. Has the following keys:
            "user_message" - A ChatMessage dict representing the user input to the chat stream
            "chat_context" - A ChatContext dict representing the state of the previous conversation
                            (if any)
        :return: An iterator of dictionary versions of the ChatResponse
                    protobufs structure. Has the following keys:
            "response"      - An optional ChatMessage dictionary.  See chat.proto for details.

            Note that responses to the chat input might be numerous and will come as they
            are produced until the system decides there are no more messages to be sent.
        """
        separator: bytes = b"\n"
        max_chunk_size: int = 64 * 1024
        path: str = self.get_request_path("streaming_chat")
        try:
            timeout: ClientTimeout = None
            if self.streaming_timeout_in_seconds is not None:
                timeout = ClientTimeout(self.streaming_timeout_in_seconds)
            async with ClientSession(headers=self.get_headers(),
                                     timeout=timeout
                                     ) as session:
                async with session.post(path, json=request_dict) as response:
                    # Check for successful response status
                    response.raise_for_status()

                    # Iterate over the content stream as it comes in.
                    # Note: We used to iterate over lines with the simpler:
                    #           async for line in response.content:
                    #               ... blah blah ...
                    #       but that could fail with ValueError("Chunk too big")
                    #       if a single line was too long.
                    accumulator: bytes = b""
                    async for data in response.content.iter_chunked(max_chunk_size):

                        # Concatenate data as it comes in
                        accumulator += data

                        # Try to find our line separator
                        index: int = accumulator.find(separator)
                        while index >= 0:

                            # Grab a single line
                            line: bytes = accumulator[:index]
                            unicode_line = line.decode("utf-8")
                            if unicode_line.strip():    # Skip empty lines

                                # We have a line with something in it.
                                # Decode and yield as a dictionary
                                result_dict = json.loads(unicode_line)
                                yield result_dict

                            # Remove the previous line from the accumulator
                            accumulator = accumulator[index + len(separator):]

                            # Allow for case of multiple lines in one chunk
                            index = accumulator.find(separator)

                    # If there is anything left in the accumulator, yield it
                    if len(accumulator) > 0:
                        result_dict = json.loads(accumulator.decode("utf-8"))
                        yield result_dict

        except (asyncio.TimeoutError, ClientOSError, ClientPayloadError) as exc:
            # Pass on a couple of asserts that are known to represent
            # real problems that a client has to deal with.
            # We figure this is OK for streaming_chat() because normally
            # in order to get to using streaming_chat() clients will most
            # often call function() first, and that will have the blanket
            # helpful asserts for the newly initiated.
            raise exc

        except Exception as exc:  # pylint: disable=broad-exception-caught
            # Assume the newly initiated need some more help.
            raise ValueError(self.help_message(path)) from exc




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict


class StructureParser:
    """
    Interface that allows parsing a dictionary from within LLM response text.
    Concrete subclasses implement parsing from various formats.

    The notion of a "remainder" allows for communicating other descriptive text
    outside of what was parsed for the dictionary itself.
    """

    def __init__(self):
        """
        Constructor
        """
        self.remainder: str = None

    def parse_structure(self, content: str) -> Dict[str, Any]:
        """
        Parse the single string content for any signs of structure

        :param content: The string to parse for structure
        :return: A dictionary structure that was embedded in the content.
                Will return None if no parseable structure is detected.
        """
        raise NotImplementedError

    def get_remainder(self) -> str:
        """
        :return: Any content string that was not essential in detecting or
                 describing the structure.  The parse_structure() method must
                 be called to get anything valid out of the return value here.
                 Will return None if no parseable structure is detected.
        """
        return self.remainder




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage

from leaf_common.config.dictionary_overlay import DictionaryOverlay

from neuro_san.internals.graph.activations.abstract_callable_activation import AbstractCallableActivation
from neuro_san.internals.graph.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.graph.interfaces.callable_activation import CallableActivation
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.run_context.factory.run_context_factory import RunContextFactory
from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.internals.run_context.interfaces.tool_call import ToolCall
from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing


class CallingActivation(AbstractCallableActivation, ToolCaller):
    """
    An implementation of the ToolCaller interface which actually does
    the calling of the other tools.

    Worth noting that this is used as a base implementation for:
        * BranchActivation - which can both call other tools and be called as a tool
        * FrontManActivation - which can only call other tools but has other specialized
            logic for interacting with user input, it being the root node of a tool graph.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self, parent_run_context: RunContext,
                 factory: AgentToolFactory,
                 agent_tool_spec: Dict[str, Any],
                 sly_data: Dict[str, Any]):
        """
        Constructor

        :param parent_run_context: The parent RunContext (if any) to pass
                             down its resources to a new RunContext created by
                             this call.
        :param factory: The factory for Agent Tools.
        :param agent_tool_spec: The dictionary describing the JSON agent tool
                            to be used by the instance
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        """
        super().__init__(factory, agent_tool_spec, sly_data)

        # Get the llm config as a combination of defaults from different places in the config
        agent_network_config: Dict[str, Any] = self.factory.get_config()
        spec_llm_config: Dict[str, Any] = self.agent_tool_spec.get("llm_config")
        run_context_config: Dict[str, Any] = self.prepare_run_context_config(agent_network_config,
                                                                             spec_llm_config)
        self.run_context: RunContext = RunContextFactory.create_run_context(parent_run_context, self,
                                                                            config=run_context_config)
        self.journal: Journal = self.run_context.get_journal()

    @staticmethod
    def prepare_run_context_config(agent_network_config: Dict[str, Any],
                                   spec_llm_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get the llm config as a combination of defaults from different places in the config

        :param agent_network_config: The entirety of the agent network's config
        :param spec_llm_config: The llm config for the agent spec
        :return: A merged llm config to use for a RunContext
        """
        empty: Dict[str, Any] = {}
        if spec_llm_config is None:
            spec_llm_config = empty

        overlayer = DictionaryOverlay()
        llm_config = agent_network_config.get("llm_config", empty)
        llm_config = overlayer.overlay(llm_config, spec_llm_config)

        run_context_config: Dict[str, Any] = {
            "context_type": agent_network_config.get("context_type"),
            "llm_config": llm_config
        }
        return run_context_config

    def get_instructions(self) -> str:
        """
        :return: The string prompt for framing the problem in terms of purpose.
        """
        instructions: str = self.agent_tool_spec.get("instructions")
        if instructions is None:
            agent_name: str = self.get_name()
            message: str = f"""
The agent named "{agent_name}" has no instructions specified for it.

Every agent must have instructions providing the natural-language
context with which it will proces input, essentially telling it what to do.
"""
            raise ValueError(message)
        return instructions

    async def create_resources(self, component_name: str = None,
                               instructions: str = None,
                               assignments: str = ""):
        """
        Creates resources that will be used throughout the lifetime of the component.
        :param component_name: Optional string for labelling the component.
                        Defaults to the agent name if not set.
        :param instructions: Optional string for setting more fine-grained instructions.
                        Defaults to agent instructions if not set.
        :param assignments: Optional string for assigning agent functional arguments.
                        Defaults to an empty string if not set.
        """
        name = component_name
        if name is None:
            name = self.get_name()

        use_instructions = instructions
        if instructions is None:
            use_instructions = self.get_instructions()

        tool_names: List[str] = self.get_callable_tool_names(self.agent_tool_spec)
        await self.run_context.create_resources(name, use_instructions, assignments, tool_names=tool_names)

    @staticmethod
    def get_callable_tool_names(agent_tool_spec: Dict[str, Any]) -> List[str]:
        """
        :return: The names of the callable tools this instance will call
                 Can return None if the instance will not call any tools.
        """
        tool_list: List[str] = agent_tool_spec.get("tools")
        if tool_list is None or len(tool_list) == 0:
            # We really don't have any callable tools
            return None

        return tool_list

    def get_callable_tool_dict(self) -> Dict[str, str]:
        """
        :return: A dictionary of name -> name, where the keys are what the tool has been
                called in the framework (e.g. langchain) and the value is what the tool
                has been called in the agent_spec itself.  This is often a reflexive
                mapping, except in the case of external tools where a safer name is
                needed for internal framework reference.
        """
        tool_dict: Dict[str, str] = {}

        tool_list: List[str] = self.get_callable_tool_names(self.agent_tool_spec)
        if tool_list is None:
            return tool_dict

        for tool in tool_list:
            safe_name: str = ExternalAgentParsing.get_safe_agent_name(tool)
            tool_dict[safe_name] = tool

        return tool_dict

    async def make_tool_function_calls(self, component_run: Run) -> Run:
        """
        Calls all of the callable_components' functions

        :param component_run: The Run which the component is operating under
        :return: A potentially updated Run for the component
        """
        # DEF: Still a mystery as to how to get langchain implementation
        #      to tell us the tool calls it *needs* to make vs the tool calls
        #      it *could* make.
        component_tool_calls: List[ToolCall] = component_run.get_tool_calls()
        tool_outputs: List[Dict[str, Any]] = []  # Initialize an empty list to store tool outputs

        # Call each of the the listed tools and collect the results
        # of their function(s).
        for component_tool_call in component_tool_calls:

            tool_output: Dict[str, Any] = await self.make_one_tool_function_call(component_tool_call)

            # Add the tool output for the current component_tool_call to the list
            tool_outputs.append(tool_output)

        # Submit all tool outputs at once after the loop has gathered all
        # outputs of all CallableActivation' functions.
        component_run = await self.run_context.submit_tool_outputs(component_run, tool_outputs)

        return component_run

    async def make_one_tool_function_call(self, component_tool_call: ToolCall) -> Dict[str, Any]:
        """
        Calls a single callable_component's function

        :param component_tool_call: A ToolCall instance to get the function
                            arguments from
        :return: A dictionary with keys:
                "tool_call_id" a string id representing the call to the tool itself
                "output" a JSON string representing the output of the tool's function
        """
        # Get the function args as a dictionary
        tool_name: str = component_tool_call.get_function_name()
        tool_arguments: Dict[str, Any] = component_tool_call.get_function_arguments()

        # Create a new instance of a JSON-speced tool using the supplied callable_tool_name.
        # At this point tool_name might be an internal reference to an external tool,
        # so we need to check a mapping.
        callable_tool_dict: Dict[str, str] = self.get_callable_tool_dict()
        use_tool_name: str = callable_tool_dict.get(tool_name)
        if use_tool_name is None:
            raise ValueError(f"{tool_name} is not in tools {list(callable_tool_dict.keys())}")

        # Note: This is not a BaseTool. This is our own construct within graph
        #       that we can build().
        our_agent_spec = self.get_agent_tool_spec()
        callable_component: CallableActivation = \
            self.factory.create_agent_activation(self.run_context, our_agent_spec, use_tool_name,
                                                 self.sly_data, tool_arguments)

        message: BaseMessage = await callable_component.build()

        # Prepare the tool output
        tool_output: Dict[str, Any] = {
            "origin": callable_component.get_origin(),
            "tool_call_id": component_tool_call.get_id(),
            "output": message,
            # Add the component's sly_data to the mix.
            # External tools have separate dictionaries of redacted sly_data that need to
            # be reintegrated with the single copy that floats around the agent network.
            "sly_data": callable_component.sly_data
        }

        # Clean up after this CallableActivation.
        # Note that the run_context passed here is used as a comparison to be sure
        # that the CallableActivation's cleanup does not accidentally clean up
        # any resources that should still remain open for this
        # CallingActivation's purposes.
        await callable_component.delete_resources(self.run_context)

        return tool_output

    async def build(self) -> BaseMessage:
        """
        Main entry point to the class.

        :return: A BaseMessage produced during this process.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Union

import logging

from ddgs import DDGS

from neuro_san.interfaces.coded_tool import CodedTool


class WebsiteSearch(CodedTool):
    """
    CodedTool implementation which provides a way to utilize different websites' search feature
    """

    def __init__(self):
        self.top_n = 5

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Union[Dict[str, Any], str]:
        """
        As much as we'd prefer an asynchronous entry point, the code below uses synchronous
        calls for a simple example. Asynchronous calls make for a more performant server
        environment.
        """
        raise NotImplementedError

    def invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Union[Dict[str, Any], str]:
        """
        :param args: An argument dictionary whose keys are the parameters
                to the coded tool and whose values are the values passed for them
                by the calling agent.  This dictionary is to be treated as read-only.

                The argument dictionary expects the following keys:
                    "app_name" the name of the One Cognizant app for which the URL is needed.

        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
                but whose values are meant to be kept out of the chat stream.

                This dictionary is largely to be treated as read-only.
                It is possible to add key/value pairs to this dict that do not
                yet exist as a bulletin board, as long as the responsibility
                for which coded_tool publishes new entries is well understood
                by the agent chain implementation and the coded_tool implementation
                adding the data is not invoke()-ed more than once.

                Keys expected for this implementation are:
                    None

        :return:
            In case of successful execution:
                The URL to the app as a string.
            otherwise:
                a text string an error message in the format:
                "Error: <error message>"
        """
        the_url: str = args.get("url", "")
        if the_url == "":
            return "Error: No URL provided."
        search_terms: str = args.get("search_terms", "")
        if search_terms == "":
            return "Error: No search terms provided."
        search_terms = the_url + " " + search_terms

        logger = logging.getLogger(self.__class__.__name__)
        logger.info(">>>>>>>>>>>>>>>>>>>WebsiteSearch>>>>>>>>>>>>>>>>>>")
        logger.info("BSearch Terms: %s", str(search_terms))
        the_links = self.search_web(search_terms, self.top_n)
        links_str = ""
        for index, the_link in enumerate(the_links, start=1):
            links_str += f"{index}. {the_link} ; "
            logger.info("%s. %s", str(index), str(the_link))
        logger.info(">>>>>>>>>>>>>>>>>>>DONE !!!>>>>>>>>>>>>>>>>>>")
        return links_str

    def search_web(self, query: str, num_results: int = 5) -> list:
        """
        Search the web for a given query using DuckDuckGo Search
        and return a list of result URLs.

        :param query: The search query (e.g., "10.5 white men sneakers").
        :param num_results: Number of links to retrieve (default=5).
        :return: List of hyperlink strings.
        """
        # Use duckduckgo_search to retrieve results
        search = DDGS()
        # Synchronous call below!
        # Try to use asynchronous calls for a more performant neuro-san server
        results = search.text(query, max_results=num_results)

        # Extract and return only the URLs from the returned list of dictionaries
        returned_links = [res["href"] for res in results if "href" in res]

        return returned_links




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Tuple

from asyncio import Event

from neuro_san.interfaces.reservation import Reservation
from neuro_san.interfaces.reservationist import Reservationist


class ReservationUtil:
    """
    A static utility class intended to be used from an asynchronous CodedTool that makes
    Agent (network) Reservations easier.
    """

    @staticmethod
    async def wait_for_one(args: Dict[str, Any], agent_spec: Dict[str, Any], lifetime_in_seconds: float,
                           prefix: str = "") \
            -> Tuple[Reservation, str]:
        """
        Waits for a single agent to be deployed as a temporary agent network.

        :param args:  The args for the CodedTool
        :param agent_spec: The dictionary containing the agent network specification
        :param lifetime_in_seconds: How long the temporary agent network should live, in seconds
        :param prefix: An optional prefix to attach to the generated reservation id.
        :return: A tuple containing the Reservation representing the agent network that was deployed
                and a string representing an error message pertaining to the Reservation.  One
                of the elements of the Tuple will be None.
        """
        reservation: Reservation = None
        error: str = None

        reservationist: Reservationist = args.get("reservationist")
        if reservationist is None:
            error = """
Reservationist is None.  Try this for your server:
    export AGENT_TEMPORARY_NETWORK_UPDATE_PERIOD_SECONDS=5
"""
            return (reservation, error)

        # Creating the Reservations can be done outside the Reservationist with-statement
        reservation: Reservation = await reservationist.reserve(lifetime_in_seconds=lifetime_in_seconds,
                                                                prefix=prefix)
        deployments: Dict[Reservation, Dict[str, Any]] = {
            reservation: agent_spec
        }

        # Deploy the reservations with confirmation event
        # If you don't really need to wait until the new agent(s) has been deployed
        # then set confirmation=False, and don't bother about waiting for the Event.
        deployed_event: Event = None
        try:
            async with reservationist:
                deployed_event = await reservationist.deploy(deployments, confirmation=True)

        except ValueError as exception:
            # Report exceptions from below as errors here.
            error = f"{exception}"

        if deployed_event is not None:
            await deployed_event.wait()

        return (reservation, error)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Set

from logging import getLogger
from logging import Logger

from neuro_san.internals.validation.network.abstract_network_validator import AbstractNetworkValidator
from neuro_san.internals.validation.network.graph_visitation_state import GraphVisitationState


class CyclesNetworkValidator(AbstractNetworkValidator):
    """
    AbstractNetworkValidator that looks for cycles in the agent graph.
    This is not strictly forbidden by neuro-san infrastructure, but there
    are some situtations where it is good to at least flag it.
    """

    def __init__(self):
        """
        Constructor
        """
        self.logger: Logger = getLogger(self.__class__.__name__)

    def validate_name_to_spec_dict(self, name_to_spec: Dict[str, Any]) -> List[str]:
        """
        Validate the agent network, specifically in the form of a name -> agent spec dictionary.

        :param name_to_spec: The name -> agent spec dictionary to validate
        :return: A list of error messages
        """
        errors: List[str] = []

        # Find cyclical agents
        cyclical_agents: Set[str] = self.find_cyclical_agents(name_to_spec)
        if cyclical_agents:
            errors.append(f"Cyclical dependencies found in agents: {sorted(cyclical_agents)}")

        if len(errors) > 0:
            # Only warn if there is a problem
            self.logger.warning(str(errors))

        return errors

    def find_cyclical_agents(self, name_to_spec: Dict[str, Any]) -> Set[str]:
        """
        Find agents that are part of cyclical dependencies using DFS.

        :param name_to_spec: The agent network to validate
        :return: Set of agent names that are part of cycles
        """
        # Step 1: Initialize state tracking for all agents
        state: Dict[str, GraphVisitationState] = {}
        for agent in name_to_spec.keys():
            state[agent] = GraphVisitationState.UNVISITED

        # Step 2: Set to collect all agents that are part of cycles
        cyclical_agents: Set[str] = set()

        # Step 3: Start DFS from each unvisited agent to ensure we check all components
        # (the network might have disconnected parts)
        for agent in name_to_spec.keys():
            if state[agent] == GraphVisitationState.UNVISITED:  # Only start DFS from unvisited agents
                # Start DFS with empty path - this agent is the root of this search
                self.dfs_cycle_detection(name_to_spec, agent, [], state, cyclical_agents)

        # Step 4: Return all agents that were found to be part of cycles
        return cyclical_agents

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def dfs_cycle_detection(self, name_to_spec: Dict[str, Any], agent: str,
                            path: List[str], state: Dict[str, GraphVisitationState],
                            cyclical_agents: Set[str]):
        """
        Perform Depth-First Search (DFS) traversal to detect cycles starting from a specific agent.

        :param name_to_spec: The agent network to validate
        :param agent: Current agent being visited
        :param path: Current path from start to current agent (for cycle identification)
        :param state: Dictionary tracking visit state of all agents
        :param cyclical_agents: Set to collect all agents that are part of any cycle
        """
        # Step 1: Check if we've encountered an agent currently being processed (back edge = cycle)
        if state[agent] == GraphVisitationState.CURRENTLY_BEING_PROCESSED:
            # Cycle detected! The agent is already in our current processing path
            cycle_start_idx: int = path.index(agent)  # Find where the cycle starts in our path
            cycle_agents: Set[str] = set(path[cycle_start_idx:] + [agent])  # Extract all agents in the cycle
            cyclical_agents.update(cycle_agents)  # Add them to our result set
            return

        # Step 2: Skip if this agent was already fully processed in a previous DFS
        if state[agent] == GraphVisitationState.FULLY_PROCESSED:
            return  # Already completed, no need to process again

        # Step 3: Mark agent as currently being processed (prevents infinite recursion)
        state[agent] = GraphVisitationState.CURRENTLY_BEING_PROCESSED

        # Step 4: Add current agent to the path (to track the route we took to get here)
        path.append(agent)

        # Step 5: Get all child agents (down_chains) of current agent
        agent_spec: Dict[str, Any] = name_to_spec.get(agent, {})
        down_chains: List[str] = agent_spec.get("tools", [])
        safe_down_chains: List[str] = self.remove_dictionary_tools(down_chains)

        # Step 6: Recursively visit each child agent
        for child_agent in safe_down_chains:
            # Only visit child if it exists in our network (safety check)
            if child_agent in name_to_spec:
                self.dfs_cycle_detection(name_to_spec, child_agent, path, state, cyclical_agents)

        # Step 7: Backtrack - remove current agent from path as we're done processing it
        path.pop()

        # Step 8: Mark agent as fully processed (all its descendants have been explored)
        state[agent] = GraphVisitationState.FULLY_PROCESSED




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from pathlib import Path

import json

from pyparsing.exceptions import ParseException
from pyparsing.exceptions import ParseSyntaxException

from leaf_common.config.config_filter_chain import ConfigFilterChain
from leaf_common.persistence.easy.easy_hocon_persistence import EasyHoconPersistence
from leaf_common.persistence.interface.restorer import Restorer

from neuro_san.internals.interfaces.agent_name_mapper import AgentNameMapper
from neuro_san.internals.graph.persistence.agent_filetree_mapper import AgentFileTreeMapper
from neuro_san.internals.graph.persistence.agent_standalone_mapper import AgentStandaloneMapper
from neuro_san.internals.graph.filters.defaults_config_filter import DefaultsConfigFilter
from neuro_san.internals.graph.filters.dictionary_common_defs_config_filter \
    import DictionaryCommonDefsConfigFilter
from neuro_san.internals.graph.filters.name_correction_config_filter import NameCorrectionConfigFilter
from neuro_san.internals.graph.filters.string_common_defs_config_filter \
    import StringCommonDefsConfigFilter
from neuro_san.internals.graph.registry.agent_network import AgentNetwork


class AgentNetworkRestorer(Restorer):
    """
    Implementation of the Restorer interface to read in an AgentNetwork
    instance given a JSON file name.
    """

    def __init__(self, registry_dir: str = None, agent_mapper: AgentNameMapper = None):
        """
        Constructor

        :param registry_dir: The directory under which file_references
                    for registry files are allowed to be found.
                    If None, there are no limits, but paths must be absolute
        :param agent_mapper: optional AgentNameMapper;
            if None, default will be used:
                if registry_dir is None, AgentStandaloneMapper instance will be used;
                otherwise, we use AgentFileTreeMapper.
        """
        self.registry_dir: str = registry_dir
        self.agent_mapper = agent_mapper
        if not self.agent_mapper:
            if self.registry_dir is not None:
                self.agent_mapper = AgentFileTreeMapper()
            else:
                self.agent_mapper = AgentStandaloneMapper()

    def restore(self, file_reference: str = None):
        """
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: an object from some persisted store
        """
        config: Dict[str, Any] = None

        if file_reference is None or len(file_reference) == 0:
            raise ValueError(f"file_reference {file_reference} cannot be None or empty string")

        use_file: str = file_reference
        if self.registry_dir is not None:
            # This should be OS-agnostic operation, producing a valid local file path
            use_file = str(Path(self.registry_dir) / file_reference)

        try:
            if use_file.endswith(".json"):
                config = json.load(use_file)
            elif use_file.endswith(".hocon"):
                hocon = EasyHoconPersistence(full_ref=use_file, must_exist=True)
                config = hocon.restore()
            else:
                raise ValueError(f"file_reference {use_file} must be a .json or .hocon file")
        except (ParseException, ParseSyntaxException, json.decoder.JSONDecodeError) as exception:
            message = f"""
There was an error parsing the agent network file "{use_file}".
See the accompanying ParseException (above) for clues as to what might be
syntactically incorrect in that file.
"""
            raise ParseException(message) from exception

        # Now create the AgentNetwork
        # Inside here is incorrectly flagged as destination of Path Traversal 7
        #   Reason: The lines above ensure that the path of registry_dir is within
        #           this source base. CheckMarx does not recognize
        #           the calls to Pathlib/__file__ as a valid means to resolve
        #           these kinds of issues.
        name = self.agent_mapper.filepath_to_agent_network_name(file_reference)
        agent_network: AgentNetwork = self.restore_from_config(name, config)
        return agent_network

    def restore_from_config(self, agent_name: str, config: Dict[str, Any]) -> AgentNetwork:
        """
        :param agent_name: name of an agent;
        :param config: agent configuration dictionary,
            built or parsed from external sources;
        :return: AgentNetwork instance for an agent.
        """
        # Perform a filter chain on the config that was read in
        filter_chain = ConfigFilterChain()
        filter_chain.register(DictionaryCommonDefsConfigFilter())
        filter_chain.register(StringCommonDefsConfigFilter())
        filter_chain.register(DefaultsConfigFilter())
        filter_chain.register(NameCorrectionConfigFilter())
        config = filter_chain.filter_config(config)

        # Now create the AgentNetwork
        agent_network = AgentNetwork(config, agent_name)

        return agent_network




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

import uuid

from neuro_san.internals.run_context.interfaces.tool_call import ToolCall


class LangChainToolCall(ToolCall):
    """
    A LangChain implementation of a ToolCall

    For the uninitiated: A "ToolCall" in langchain/openai parlance is a *request*
    that a tool be called with certain structured function arguments.
    """

    def __init__(self, tool_name: str, args: Any, run_id: str):
        """
        Constructor

        :param tool_name: The name of the tool to be called
        :param args: The arguments the tool is requested to be called with
                So far we've only seen this as Dict[str, Any], but the langchain
                typing is Any, so we stick with that.
        :param run_id: The string id of the parent run so that the tool's
                    ids can be associated with that.
        """
        self.tool_name: str = tool_name
        self.args = args
        self.id: str = f"tool_call_{run_id}_{uuid.uuid4()}"

    def get_id(self) -> str:
        """
        :return: The string id of this run
        """
        return self.id

    def get_function_arguments(self) -> Dict[str, Any]:
        """
        :return: Returns a dictionary of the function arguments for the tool call
        """
        return self.args

    def get_function_name(self) -> str:
        """
        :return: Returns the string name of the tool
        """
        return self.tool_name




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT


class AgentNameMapper:
    """
    An abstract policy defining conversion
    between agent name as specified in a manifest file
    and a file path (relative to registry root directory) to this agent definition file.
    """

    def agent_name_to_filepath(self, agent_name: str) -> str:
        """
        Converts an agent name from manifest file to file path to this agent definition file.
        """
        raise NotImplementedError()

    def filepath_to_agent_network_name(self, filepath: str) -> str:
        """
        Converts a file path to agent definition file (relative to registry root directory)
        to agent network name identifying it to the service.
        """
        raise NotImplementedError()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.service.watcher.interfaces.startable import Startable


class StorageUpdater(Startable):
    """
    Interface for specific updating jobs that the Watcher performs.
    """

    def start(self):
        """
        Perform start up.
        """
        raise NotImplementedError

    def get_update_period_in_seconds(self) -> int:
        """
        :return: An int describing how long this updater ideally wants to go between
                calls to update_storage().
        """
        raise NotImplementedError

    def needs_updating(self, time_now_in_seconds: float) -> bool:
        """
        :param time_now_in_seconds: The current time in seconds.
                    We expect this to be from time.time()
        :return: True if this instance needs updating. False otherwise.
        """
        raise NotImplementedError

    def update_storage(self):
        """
        Perform an update
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.interfaces.agent_progress_reporter import AgentProgressReporter
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.agent_progress_message import AgentProgressMessage


class ProgressJournal(AgentProgressReporter):
    """
    An implementation of the AgentProgressReporter interface for a CodedTool to be able
    to journal AgentProgressMessages.
    """

    def __init__(self, wrapped_journal: Journal):
        """
        Constructor

        :param wrapped_journal: The Journal that this implementation wraps
        """
        self.wrapped_journal: Journal = wrapped_journal

    async def async_report_progress(self, structure: Dict[str, Any], content: str = ""):
        """
        Reports the structure and optional message to the chat message stream returned to the client
        To be used from within CodedTool.async_invoke().

        :param structure: The Dictionary instance to write as progress.
                        All keys and values must be JSON-serializable.
        :param content: An optional message to send to the client
        """
        if structure is None:
            # Nothing to report
            return
        if not isinstance(structure, Dict):
            raise ValueError(f"Expected dictionary, got {type(structure)}")

        if content is None:
            content = ""

        message = AgentProgressMessage(content, structure=structure)
        await self.wrapped_journal.write_message(message)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class definition for comments.
"""
import os
import sys
import stat
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
import psutil

# Unix-only
try:
    # pylint: disable=invalid-name
    import resource  # not available on Windows
except Exception:  # pylint: disable=broad-exception-caught
    resource = None


class ServiceResources:
    """
    Class provides utility methods to monitor usage
    of service run-time resources,
    like:
    - Unix/macOS: file descriptors (FDs) by type
    - Windows: OS handles breakdown (files vs INET sockets) + total handles
    """
    on_unix: bool = sys.platform.startswith("linux")
    on_macos: bool = sys.platform.startswith("darwin")
    on_windows: bool = sys.platform.startswith("win")

    # ---------------------------
    # POSIX helpers (Linux/macOS)
    # ---------------------------
    @classmethod
    def _iter_fds_posix(cls):
        """
        Iterator over numeric FDs for the current process (Unix/macOS).
        """
        fd_dir = "/proc/self/fd" if cls.on_unix else "/dev/fd"
        try:
            names = os.listdir(fd_dir)
        except Exception:  # pylint: disable=broad-exception-caught
            # directory may not exist in some rare environments
            return
        for name in names:
            try:
                fd = int(name)
            except Exception:  # pylint: disable=broad-exception-caught
                continue
            yield fd

    @classmethod
    def _classify_fds_posix(cls) -> Dict[str, int]:
        """
        Returns counts by FD kind on Unix/macOS:
          regular_file, socket_inet, socket_unix, fifo_pipe, other, total
        """
        p = psutil.Process()

        # Maps of socket FDs to distinguish AF_INET vs AF_UNIX
        inet_fds = {c.fd for c in p.connections(kind="inet")}  # tcp/udp
        unix_fds = {c.fd for c in p.connections(kind="unix")}  # unix sockets

        fd_dict: Dict[str, int] = {}
        total_fds = 0

        for fd in cls._iter_fds_posix() or ():
            try:
                st = os.fstat(fd)
            except OSError:
                continue  # fd may have just closed
            mode = st.st_mode

            if stat.S_ISREG(mode):
                kind = "regular_file"
            elif stat.S_ISSOCK(mode):
                if fd in inet_fds:
                    kind = "socket_inet"
                elif fd in unix_fds:
                    kind = "socket_unix"
                else:
                    kind = "other"  # socket but not recognized by psutil maps
            elif stat.S_ISFIFO(mode):
                kind = "fifo_pipe"
            else:
                kind = "other"

            total_fds += 1
            fd_dict[kind] = fd_dict.get(kind, 0) + 1

        fd_dict["total"] = total_fds
        # ensure all expected keys exist (helpful for stable metrics)
        for k in ("regular_file", "socket_inet", "socket_unix", "fifo_pipe", "other"):
            fd_dict.setdefault(k, 0)
        return fd_dict

    # ---------------------------
    # Windows helpers
    # ---------------------------
    @classmethod
    def _classify_handles_windows(cls) -> Dict[str, int]:
        """
        Returns a simplified breakdown on Windows using psutil:

          regular_file   -> len(Process.open_files())
          socket_inet    -> len(Process.connections(kind="inet"))
          other_handles  -> num_handles - above_known
          total_handles  -> Process.num_handles()

        Notes:
          * Windows does not expose POSIX FDs; we count OS handles instead.
          * We cannot reliably enumerate every handle type without native WinAPI.
        """
        p = psutil.Process()

        try:
            total_handles = p.num_handles()  # all handles owned by this process
        except Exception:  # pylint: disable=broad-exception-caught
            total_handles = 0
        files = len(p.open_files())
        # TCP/UDP INET sockets for this process
        inet_conns = p.connections(kind="inet")
        socket_inet = len(inet_conns)

        # We can't see AF_UNIX on Windows (not applicable), FIFO pipes classification requires WinAPI.
        # Derive "other_handles" as the remainder; never let it go negative.
        known = files + socket_inet
        other_handles = max(total_handles - known, 0)

        return {
            "regular_file": files,
            "socket_inet": socket_inet,
            "socket_unix": 0,         # not applicable on Windows
            "fifo_pipe": 0,           # not available via psutil alone
            "other_handles": other_handles,
            "total_handles": total_handles,
        }

    # ---------------------------
    # Public API (cross-platform)
    # ---------------------------
    @classmethod
    def classify_fds(cls) -> Dict[str, int]:
        """
        Cross-platform classification:
          * Unix/macOS: returns per-FD kinds + "total"
          * Windows:    returns per-handle kinds + "total_handles"
        """
        if cls.on_unix or cls.on_macos:
            return cls._classify_fds_posix()
        if cls.on_windows:
            return cls._classify_handles_windows()
        # Fallback: try POSIX path; if not, return minimal info
        try:
            return cls._classify_fds_posix()
        except Exception:  # pylint: disable=broad-exception-caught
            p = psutil.Process()
            return {"total_unknown": getattr(p, "num_fds", lambda: 0)()}

    @classmethod
    def get_fd_usage(cls) -> Tuple[Dict[str, int], Optional[int], Optional[int]]:
        """
        Returns (counts_dict, soft_limit, hard_limit).

        * Unix/macOS: soft/hard are RLIMIT_NOFILE integers.
        * Windows:    returns (counts_dict, None, None) since RLIMIT_NOFILE does not apply.
        """
        counts = cls.classify_fds()

        if (cls.on_unix or cls.on_macos) and resource is not None:
            soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
            return counts, soft, hard

        # Windows / unknown
        return counts, None, None

    # --- internal helper: get this process's TCP connections, psutil-6-safe ---
    @classmethod
    def _proc_tcp_conns(cls) -> List[Any]:
        """
        Return TCP connections for the current process, with a fallback that
        is compatible with psutil 6.x where Process.connections may be deprecated.
        """
        p = psutil.Process()
        try:
            # Works on psutil <= 5.x and (for now) also on many 6.x installs
            return p.connections(kind="tcp")
        except Exception:  # pylint: disable=broad-exception-caught
            # psutil 6.x style: filter system-wide by pid
            conns = []
            for c in psutil.net_connections(kind="tcp"):
                if getattr(c, "pid", None) == p.pid:
                    conns.append(c)
            return conns

    # --- unchanged external API, but hardened for Windows/psutil variations ---
    @classmethod
    def classify_outbound_sockets(cls, outbound_tcp: Iterable[Any]) -> Dict[str, Any]:
        """
        Classify outbound process socket connections by remote address and status.

        Cross-platform notes:
          * Windows/macOS/Linux all provide per-process TCP connections via psutil.
          * Some entries may have missing raddr (e.g., early handshake) — we bucket those under "(no-remote)".
          * Status names are strings from psutil (e.g., 'ESTABLISHED','TIME_WAIT', etc.).
        """
        result: Dict[str, Dict[str, int]] = {}

        for conn in outbound_tcp:
            # Normalize remote address: "ip:port" or "(no-remote)"
            if conn.raddr:
                try:
                    rip = getattr(conn.raddr, "ip", None)
                    rport = getattr(conn.raddr, "port", None)
                    if rip is None or rport is None:
                        # Fallback to repr if platform provides a different shape
                        sock_addr = str(conn.raddr)
                    else:
                        sock_addr = f"{rip}:{rport}"
                except Exception:  # pylint: disable=broad-exception-caught
                    # Very defensive: some psutil versions/platforms might differ
                    sock_addr = str(conn.raddr)
            else:
                sock_addr = "(no-remote)"

            bucket = result.setdefault(sock_addr, {})
            sock_status: str = str(getattr(conn, "status", "UNKNOWN"))
            bucket[sock_status] = bucket.get(sock_status, 0) + 1

        return result

    @classmethod
    def classify_sockets(cls, server_port: int) -> Dict[str, Any]:
        """
        Classify active sockets bound to the given server port.
        :param server_port: server port
        :return: dictionary with keys:
           "inbound_listen": number of inbound listening sockets;
           "inbound_accepted": number of accepted inbound connections;
           "outbound_tcp": dictionary describing outbound connections.

        Cross-platform behavior:
          * Works on Linux/macOS/Windows using per-process TCP enumeration.
          * Dual-stack listeners typically appear as separate IPv4/IPv6 LISTEN sockets.
          * Multi-process servers (Tornado server.start(N)) must call this in each worker PID.
        """
        tcp = cls._proc_tcp_conns()

        inbound_listen = 0
        inbound_accepted_list: list = []
        outbound_list: list = []

        for c in tcp:
            lport = c.laddr.port if c.laddr else None
            status = getattr(c, "status", None)

            # LISTEN socket on our server port?
            if status == psutil.CONN_LISTEN and lport == server_port:
                inbound_listen += 1
                continue

            # Accepted inbound socket uses the server port as local port
            if lport == server_port and status != psutil.CONN_LISTEN:
                inbound_accepted_list.append(c)
                continue

            # Everything else is an outbound (client) socket from this process
            outbound_list.append(c)

        return {
            "inbound_listen": inbound_listen,
            "inbound_accepted": len(inbound_accepted_list),
            "outbound_tcp": cls.classify_outbound_sockets(outbound_list),
        }




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

import re
from re import Match
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from json.decoder import JSONDecodeError
from json_repair import loads

from neuro_san.internals.parsers.structure.structure_parser import StructureParser


class JsonStructureParser(StructureParser):
    """
    JSON implementation for a StructureParser.
    """

    def parse_structure(self, content: str) -> Dict[str, Any]:
        """
        Parse the single string content for any signs of structure

        :param content: The string to parse for structure
        :return: A dictionary structure that was embedded in the content.
                Will return None if no parseable structure is detected.
        """
        # Reset remainder on each call
        self.remainder = None

        meat: str = content
        delimiters: Dict[str, str] = {
            # Start : End
            "```json": "```",
            "```": "```",
            "`{": "}`",
            "{": "}",
        }

        meat, self.remainder = self._extract_delimited_block(content, delimiters)

        # Attempt parsing the structure from the meat
        structure: Dict[str, Any] = None

        try:
            structure = loads(meat)
            if not isinstance(structure, Dict):
                # json_repair seems to sometimes return an empty string if there is nothing
                # for it to grab onto.
                structure = None
        except JSONDecodeError:
            # Couldn't parse
            self.remainder = None
        except TypeError:
            # meat is None
            self.remainder = None

        return structure

    def _extract_delimited_block(self, text: str, delimiters: Dict[str, str]) -> Tuple[Optional[str], str]:
        """
        Extracts a block of text from the input string "text" that is enclosed between any
        of the provided delimiter pairs. Returns a tuple of:
            - The extracted main block with delimiters, or None if no match
            - The remaining string with the block removed and extra whitespace collapsed

        :param text: The input string potentially containing a delimited block
        :param delimiters: A dictionary mapping starting delimiters to ending delimiters

        :return: A tuple of (main block content, remainder string)
        """
        # Try each delimiter pair in order
        for start, end in delimiters.items():
            # Build a regex pattern to find content between start and end delimiters
            # - re.escape ensures special characters like "{" are treated literally
            # - (.*) is a greedy match for any characters between the delimiters
            pattern: str = re.escape(start) + r"(.*)" + re.escape(end)

            # Perform regex search across multiple lines if needed (DOTALL allows "." to match newlines)
            match: Match[str] = re.search(pattern, text, re.DOTALL)

            if match:
                # Extract the matched content (including the delimiters), removing leading/trailing whitespace
                main: str = match.group(0).strip()

                # Remove the matched block (including delimiters) from the input string
                remainder: str = text[:match.start()] + text[match.end():]

                return main, remainder.strip()

        # If no matching delimiters were found, return None and the full cleaned-up input
        return None, text.strip()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider


class AgentStorageSource:
    """
    Interface onto AgentNetworkStorage so that there is not a tangle with the service side.
    """

    def get_agent_network_provider(self, agent_name: str) -> AgentNetworkProvider:
        """
        Get AgentNetworkProvider for a specific agent
        :param agent_name: name of an agent
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from neuro_san.client.agent_session_factory import AgentSessionFactory
from neuro_san.client.streaming_input_processor import StreamingInputProcessor
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.message_processing.basic_message_processor import BasicMessageProcessor


class SimpleOneShot:
    """
    Encapulates basic operation of sending a single message to an agent
    and getting single aswer back.
    """

    def __init__(self, agent: str, connection_type: str = "direct",
                 host: str = None, port: int = None):
        """
        Constructor

        :param agent: The name of the agent to talk to
        :param connection_type: The string type of connection.
                    Can be http, grpc, or direct (for library usage).
                    Default is direct.
        :param hostname: The name of the host to connect to (if applicable)
        :param port: The port on the host to connect to (if applicable)
        """
        self.agent: str = agent
        self.connection_type: str = connection_type
        self.host: str = host
        self.port: int = port

    def get_answer_for(self, text: str) -> str:
        """
        Sends text to the agent and returns the agent's answer
        :param text: The text to send to the agent
        :return: The text answer from the agent.
        """
        session: AgentSession = AgentSessionFactory().create_session(self.connection_type,
                                                                     self.agent,
                                                                     hostname=self.host,
                                                                     port=self.port)
        input_processor = StreamingInputProcessor(session=session)
        processor: BasicMessageProcessor = input_processor.get_message_processor()
        request: Dict[str, Any] = input_processor.formulate_chat_request(text)

        # Call streaming_chat()
        empty: Dict[str, Any] = {}
        for chat_response in session.streaming_chat(request):
            message: Dict[str, Any] = chat_response.get("response", empty)
            processor.process_message(message, chat_response.get("type"))

        raw_answer: str = processor.get_compiled_answer()

        return raw_answer




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.internals.interfaces.agent_storage_source import AgentStorageSource


class AgentStateListener:
    """
    Abstract interface for publishing agent state changes -
    when an agent is being added or removed from the service.
    """

    def agent_added(self, agent_name: str, source: AgentStorageSource):
        """
        Agent is being added to the service.
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        raise NotImplementedError

    def agent_modified(self, agent_name: str, source: AgentStorageSource):
        """
        Existing agent has been modified in service scope.
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        raise NotImplementedError

    def agent_removed(self, agent_name: str, source: AgentStorageSource):
        """
        Agent is being removed from the service.
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.coded_tools.math_guy.calculator import Calculator as BaseCalculator


# No need to repeat code, we just want to be sure that the deep reference to the
# coded tool can be reached.
Calculator = BaseCalculator




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from logging import getLogger
from logging import Logger
from math import gcd as greatest_common_divisor
from threading import Thread
from time import sleep
from time import time

from neuro_san.service.watcher.interfaces.startable import Startable
from neuro_san.service.watcher.interfaces.storage_updater import StorageUpdater
from neuro_san.service.watcher.registries.registry_storage_updater import RegistryStorageUpdater
from neuro_san.service.watcher.temp_networks.temp_network_storage_updater import TempNetworkStorageUpdater
from neuro_san.service.utils.server_context import ServerContext
from neuro_san.service.utils.server_status import ServerStatus


class StorageWatcher(Startable):
    """
    Class implementing periodic server updates
    by watching agent files and manifest file itself
    and other changes to AgentNetworkStorage instances.
    """

    def __init__(self,
                 watcher_config: Dict[str, Any],
                 server_context: ServerContext):
        """
        Constructor.

        :param watcher_config: A config dict for StorageUpdaters
        :param server_context: ServerContext for global-ish state
        """
        self.logger: Logger = getLogger(self.__class__.__name__)
        self.updater_thread = Thread(target=self._run, daemon=True)
        self.server_context: ServerContext = server_context

        self.storage_updaters: List[StorageUpdater] = [
            RegistryStorageUpdater(server_context.get_network_storage_dict(), watcher_config),
            TempNetworkStorageUpdater(server_context.get_network_storage_dict(), watcher_config,
                                      server_context.get_queues())
        ]

        self.update_period_in_seconds: int = self.compute_update_period_in_seconds(self.storage_updaters)

        self.keep_running: bool = True

    @staticmethod
    def compute_update_period_in_seconds(storage_updaters: List[StorageUpdater]) -> int:
        """
        :param storage_updaters: A list of StorageUpdaters each with their own updating
                    timing needs.
        :return: A greatest common divisor of everything in the list, or 0
                if collectively everyone says there is nothing to do.
        """
        # Figure out what the top-level update period should be.
        update_periods: List[int] = []
        for storage_updater in storage_updaters:
            update_period: int = storage_updater.get_update_period_in_seconds()
            if update_period > 0:
                # Anybody whose update period is 0 doesn't count.
                update_periods.append(update_period)

        # If we have nobody, then we shouldn't even run.
        # Luckily greatest_common_divisor() returns 0 for an empty list.
        update_period_in_seconds: int = greatest_common_divisor(*update_periods)
        return update_period_in_seconds

    def start(self):
        """
        Start running periodic StorageUpdaters.
        """
        self.logger.info("Starting StorageWatcher with %d seconds period",
                         self.update_period_in_seconds)

        for storage_updater in self.storage_updaters:
            storage_updater.start()

        self.updater_thread.start()

    def _run(self):
        """
        Function runs manifest file update cycle.
        """
        if self.update_period_in_seconds <= 0:
            # We should not run at all.
            return

        # Initial value entering the loop
        sleep_for_seconds: float = self.update_period_in_seconds
        while self.keep_running:

            server_status: ServerStatus = self.server_context.get_server_status()
            server_status.updater.set_status(True)

            sleep(sleep_for_seconds)

            # Snap the same time so all updaters get the same sense
            # of the interval.
            time_now_in_seconds: float = time()
            for storage_updater in self.storage_updaters:
                if storage_updater.needs_updating(time_now_in_seconds):
                    storage_updater.update_storage()

            # See how long it took to update everyone.
            elapsed_time_in_seconds: float = time() - time_now_in_seconds

            # sleep() for just the right amount of time for next time around.
            # If it took longer than one cycle, then yield the cpu at least.
            sleep_for_seconds = max(self.update_period_in_seconds - elapsed_time_in_seconds, 0)

    def stop(self):
        """
        Stop running periodic StorageUpdaters.
        """
        self.logger.info("Stopping StorageWatcher with %d seconds period",
                         self.update_period_in_seconds)

        self.keep_running = False

        # Do this all in opposite order from start()

        # Wait for the thread to finish
        self.updater_thread.join()

        for storage_updater in self.storage_updaters:
            storage_updater.stop()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Generator

import json
import contextlib
import uuid

from janus import Queue

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor
from leaf_common.asyncio.asyncio_executor_pool import AsyncioExecutorPool
from leaf_common.utils.atomic_counter import AtomicCounter

from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.run_context.factory.master_toolbox_factory import MasterToolboxFactory
from neuro_san.internals.run_context.factory.master_llm_factory import MasterLlmFactory
from neuro_san.service.generic.service_agent_reservationist import ServiceAgentReservationist
from neuro_san.service.generic.agent_server_logging import AgentServerLogging
from neuro_san.service.generic.chat_message_converter import ChatMessageConverter
from neuro_san.service.interfaces.event_loop_logger import EventLoopLogger
from neuro_san.service.usage.usage_logger_factory import UsageLoggerFactory
from neuro_san.service.usage.wrapped_usage_logger import WrappedUsageLogger
from neuro_san.service.utils.server_context import ServerContext
from neuro_san.session.async_direct_agent_session import AsyncDirectAgentSession
from neuro_san.session.external_agent_session_factory import ExternalAgentSessionFactory
from neuro_san.session.session_invocation_context import SessionInvocationContext

# A list of methods to not log requests for
# Some of these can be way too chatty
DO_NOT_LOG_REQUESTS = [
]


# pylint: disable=too-many-instance-attributes
class AsyncAgentService:
    """
    A base implementation of the Neuro-San Async Agent Service,
    independent of target transport protocol.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self,
                 request_logger: EventLoopLogger,
                 security_cfg: Dict[str, Any],
                 agent_name: str,
                 agent_network_provider: AgentNetworkProvider,
                 server_logging: AgentServerLogging,
                 server_context: ServerContext):
        """
        :param request_logger: The instance of the EventLoopLogger that helps
                        log information from running event loop
        :param security_cfg: A dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  If None, uses insecure channel.
        :param agent_name: The agent name for the service
        :param agent_network_provider: The AgentNetworkProvider to use for the session.
        :param server_logging: An AgentServerLogging instance initialized so that
                        spawned asynchronous threads can also properly initialize
                        their logging.
        :param server_context: The ServerContext holding global-ish state
        """
        self.request_logger = request_logger
        self.security_cfg = security_cfg
        self.server_logging: AgentServerLogging = server_logging

        # Stuff needed for ServiceAgentReservationist
        self.queues: Queue[AsyncCollatingQueue] = server_context.get_queues()

        self.agent_network_provider: AgentNetworkProvider = agent_network_provider
        self.agent_name: str = agent_name
        self.request_counter = AtomicCounter()

        agent_network: AgentNetwork = self.agent_network_provider.get_agent_network()
        config: Dict[str, Any] = agent_network.get_config()
        self.llm_factory: ContextTypeLlmFactory = MasterLlmFactory.create_llm_factory(config)
        self.toolbox_factory: ContextTypeToolboxFactory = MasterToolboxFactory.create_toolbox_factory(config)
        self.async_executor_pool: AsyncioExecutorPool = server_context.get_executor_pool()
        # Load once.
        self.llm_factory.load()
        self.toolbox_factory.load()

    def get_request_count(self) -> int:
        """
        :return: The number of currently active requests
        """
        return self.request_counter.get_count()

    async def function(self, request_dict: Dict[str, Any],
                       request_metadata: Dict[str, Any]) \
            -> Dict[str, Any]:
        """
        Allows a client to get the outward-facing function for the agent
        served by this service.

        :param request_dict: a FunctionRequest dictionary
        :param request_metadata: request metadata
        :return: a FunctionResponse dictionary
        """
        self.request_counter.increment()
        do_log: bool = "Function" not in DO_NOT_LOG_REQUESTS
        log_marker: str = "function request"
        metadata: Dict[str, str] = {
            "request_id": f"server-{uuid.uuid4()}"
        }
        metadata.update(request_metadata)
        if do_log:
            self.request_logger.info(
                metadata,
                "Received a %s request for %s",
                f"{self.agent_name}.Function", log_marker)

        # Delegate to Direct*Session
        agent_network: AgentNetwork = self.agent_network_provider.get_agent_network()
        session: AsyncDirectAgentSession =\
            AsyncDirectAgentSession(
                agent_network=agent_network,
                invocation_context=None,
                metadata=metadata,
                security_cfg=self.security_cfg)
        response_dict = await session.function(request_dict)

        if do_log:
            self.request_logger.info(
                metadata,
                "Done with %s request for %s",
                f"{self.agent_name}.Function", log_marker)

        self.request_counter.decrement()
        return response_dict

    async def connectivity(self, request_dict: Dict[str, Any],
                           request_metadata: Dict[str, Any]) \
            -> Dict[str, Any]:
        """
        Allows a client to get connectivity information for the agent
        served by this service.

        :param request_dict: a ChatRequest dictionary
        :param request_metadata: request metadata
        :return: a ConnectivityResponse dictionary
        """
        self.request_counter.increment()
        do_log: bool = "Connectivity" not in DO_NOT_LOG_REQUESTS
        log_marker: str = "connectivity request"
        metadata: Dict[str, str] = {
            "request_id": f"server-{uuid.uuid4()}"
        }
        metadata.update(request_metadata)

        if do_log:
            self.request_logger.info(
                metadata,
                "Received a %s request for %s",
                f"{self.agent_name}.Connectivity", log_marker)

        # Delegate to Direct*Session
        agent_network: AgentNetwork = self.agent_network_provider.get_agent_network()
        session: AsyncDirectAgentSession =\
            AsyncDirectAgentSession(
                agent_network=agent_network,
                invocation_context=None,
                metadata=metadata,
                security_cfg=self.security_cfg)
        response_dict = await session.connectivity(request_dict)

        if do_log:
            self.request_logger.info(
                metadata,
                "Done with %s request for %s",
                f"{self.agent_name}.Connectivity", log_marker)

        self.request_counter.decrement()
        return response_dict

    # pylint: disable=too-many-locals
    async def streaming_chat(self, request_dict: Dict[str, Any],
                             request_metadata: Dict[str, Any]) \
            -> Generator[Dict[str, Any], None, None]:
        """
        Initiates or continues the agent chat with the session_id
        context in the request.

        :param request_dict: a ChatRequest dictionary
        :param request_metadata: request metadata
        :return: an iterator for (eventually) returned responses dictionaries
        """
        self.request_counter.increment()
        user_text: str = request_dict.get("user_message", {}).get("text", "")
        do_log: bool = "StreamingChat" not in DO_NOT_LOG_REQUESTS

        log_marker = f"'{user_text}'"
        metadata: Dict[str, str] = {
            "request_id": f"server-{uuid.uuid4()}"
        }
        metadata.update(request_metadata)

        if do_log:
            self.request_logger.info(
                metadata,
                "Received a %s request for %s",
                f"{self.agent_name}.StreamingChat", log_marker)

        # Create a reservationist for the occasion
        reservationist: Reservationist = None
        if self.queues is not None:
            reservationist = ServiceAgentReservationist()
            self.queues.sync_q.put(reservationist.get_queue())

        # Prepare
        factory = ExternalAgentSessionFactory(use_direct=False)
        invocation_context = SessionInvocationContext(
            factory,
            self.async_executor_pool,
            self.llm_factory,
            self.toolbox_factory,
            metadata,
            reservationist)
        invocation_context.start()

        # Set up logging inside async thread
        # Prefer any request_id from the client over what we generated on the server.
        executor: AsyncioExecutor = invocation_context.get_asyncio_executor()
        _ = executor.submit(None, self.server_logging.setup_logging, metadata, metadata.get("request_id"))

        # Delegate to Direct*Session
        agent_network: AgentNetwork = self.agent_network_provider.get_agent_network()
        session: AsyncDirectAgentSession =\
            AsyncDirectAgentSession(
                agent_network=agent_network,
                invocation_context=invocation_context,
                metadata=metadata,
                security_cfg=self.security_cfg)
        # Get our args in order to pass to transport-agnostic session level
        response_dict_generator: Generator[Dict[str, Any], None, None] = session.streaming_chat(request_dict)

        # See if we want to put the request dict in the response
        chat_filter_dict: Dict[str, Any] = {}
        chat_filter_dict = request_dict.get("chat_filter", chat_filter_dict)
        chat_filter_type: str = chat_filter_dict.get("chat_filter_type", "MINIMAL")

        try:
            async for response_dict in response_dict_generator:
                # Prepare chat message for output:
                response_dict = ChatMessageConverter().to_dict(response_dict)
                # Do not return the request when the filter is MINIMAL
                if chat_filter_type != "MINIMAL":
                    response_dict["request"] = request_dict
                yield response_dict
        finally:
            # Put async generator cleanup logic in "finally" part of try-except block;
            # this way we guarantee that underlying response_dict_generator will be closed
            # whether we finish consuming its data stream normally
            # OR we are interrupted downstream
            # and have special "GeneratorExit" exception delivered to us.
            request_reporting: Dict[str, Any] = invocation_context.get_request_reporting()
            # Properly close our async generator:
            if response_dict_generator is not None:
                with contextlib.suppress(Exception):
                    await response_dict_generator.aclose()
            # Ensure that our SessionInvocationContext is always closed,
            # even if generator is interrupted.
            invocation_context.close()
            invocation_context = None

        # Maybe report token accounting to a UsageLogger
        token_dict: Dict[str, Any] = request_reporting.get("token_accounting")
        if token_dict is not None:
            usage_logger: WrappedUsageLogger = UsageLoggerFactory.create_usage_logger()
            await usage_logger.log_usage(token_dict, request_metadata)

        # Iterator has finally signaled that there are no more responses to be had.
        # Log that we are done.
        if do_log:
            reporting: str = None
            if request_reporting is not None:
                reporting = json.dumps(request_reporting, indent=4, sort_keys=False)
            self.request_logger.info(metadata, "Request reporting: %s", reporting)
            self.request_logger.info(
                metadata,
                "Done with %s request for %s",
                f"{self.agent_name}.StreamingChat", log_marker)

        self.request_counter.decrement()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from asyncio import AbstractEventLoop

from copy import deepcopy
from logging import getLogger
from logging import Logger

from langchain_core.messages.ai import AIMessage
from langchain_core.messages.base import BaseMessage

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor
from leaf_common.config.resolver import Resolver
from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.interfaces.coded_tool import CodedTool
from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.graph.activations.abstract_callable_activation import AbstractCallableActivation
from neuro_san.internals.graph.activations.branch_activation import BranchActivation
from neuro_san.internals.graph.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.journals.progress_journal import ProgressJournal
from neuro_san.internals.journals.tool_argument_reporting import ToolArgumentReporting
from neuro_san.internals.messages.agent_message import AgentMessage
from neuro_san.internals.messages.origination import Origination
from neuro_san.internals.reservations.accumulating_agent_reservationist import AccumulatingAgentReservationist
from neuro_san.internals.run_context.factory.run_context_factory import RunContextFactory
from neuro_san.internals.run_context.interfaces.run_context import RunContext


class AbstractClassActivation(AbstractCallableActivation):
    """
    CallableActivation which can invoke a CodedTool by its class name.

    This is a base class for tools that dynamically invoke a Python class based on a
    fully qualified class reference. Subclasses must implement "get_full_class_ref"
    method to determine the target class.

    There are two main subclasses:
    - ClassActivation: retrieves the class reference directly from the tool specification.
    - ToolboxActivation: looks up the class reference from a predefined toolbox.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self, parent_run_context: RunContext,
                 factory: AgentToolFactory,
                 arguments: Dict[str, Any],
                 agent_tool_spec: Dict[str, Any],
                 sly_data: Dict[str, Any]):
        """
        Constructor

        :param parent_run_context: The parent RunContext (if any) to pass
                             down its resources to a new RunContext created by
                             this call.
        :param factory: The AgentToolFactory used to create tools
        :param arguments: A dictionary of the tool function arguments passed in by the LLM
        :param agent_tool_spec: The dictionary describing the JSON agent tool
                            to be used by the instance
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
                 This gets passed along as a distinct argument to the referenced python class's
                 invoke() method.
        """
        super().__init__(factory, agent_tool_spec, sly_data)
        self.run_context: RunContext = RunContextFactory.create_run_context(parent_run_context, self)
        self.journal: Journal = self.run_context.get_journal()
        self.full_name: str = Origination.get_full_name_from_origin(self.run_context.get_origin())
        self.logger: Logger = getLogger(self.full_name)

        # One of these per CodedTool
        self.reservationist: Reservationist = None

        # See if we should be doing anything with Reservationists at all
        spec_extractor = DictionaryExtractor(self.agent_tool_spec)
        if spec_extractor.get("allow.reservations"):
            invocation_context: InvocationContext = self.run_context.get_invocation_context()
            real_reservationist: Reservationist = invocation_context.get_reservationist()
            if real_reservationist is not None:
                self.reservationist = AccumulatingAgentReservationist(real_reservationist, self.full_name)

        # Put together the arguments to pass to the CodedTool
        self.arguments: Dict[str, Any] = {}
        if arguments is not None:
            self.arguments = arguments

        # Set some standard args so CodedTool can know about origin, but only if they are
        # not already set by other infrastructure.
        if self.arguments.get("origin") is None:
            self.arguments["origin"] = deepcopy(self.run_context.get_origin())
        if self.arguments.get("origin_str") is None:
            self.arguments["origin_str"] = self.full_name

        # Set some standard args that are policy objects for CodedTool consumption.
        # If you are adding keys to this section, be sure you also add to the list
        # in ToolArgumentReporting so args messages can be properly serialized.
        if self.arguments.get("progress_reporter") is None:
            self.arguments["progress_reporter"] = ProgressJournal(self.journal)
        if self.arguments.get("reservationist") is None and self.reservationist:
            # This is the Reservationist we pass into the CodedTool.
            # You might think this belongs in sly_data, but sly_data is actually a global
            # available to all CodedTools and this Reservationist is particular to the
            # CodedTool, so it goes in the arguments.
            # We specifically give the CodedTools the accumulating version so they
            # do not have any access to service internals.
            self.arguments["reservationist"] = self.reservationist

    def get_full_class_ref(self) -> str:
        """
        Returns the full class reference path of the target tool to be invoked.

        This method must be implemented by subclasses to provide the fully qualified
        class name (e.g., "my_module.MyToolClass") of the CodedTool that should
        be instantiated and invoked. This string will be used for dynamic instantiation
        of the tool class.

        :return: A dot-separated string representing the full class path.
        """
        raise NotImplementedError

    # pylint: disable=too-many-locals
    async def build(self) -> BaseMessage:
        """
        Main entry point to the class.

        :return: A BaseMessage produced during this process.
        """
        message: BaseMessage = None

        full_class_ref: str = self.get_full_class_ref()
        self.logger.info("Calling class %s", full_class_ref)
        class_split = full_class_ref.split(".")
        class_name = class_split[-1]
        # Remove the class name from the end to get the module name
        module_name = full_class_ref[:-len(class_name)]
        # Remove any trailing .s
        while module_name.endswith("."):
            module_name = module_name[:-1]

        # Resolve the class and the method
        this_agent_tool_path: str = self.factory.get_agent_tool_path()
        packages: List[str] = [this_agent_tool_path]
        resolver: Resolver = Resolver(packages)

        try:
            python_class = resolver.resolve_class_in_module(class_name, module_name)
        except (ValueError, AttributeError):
            # Drop the last segment to go one level up in the module path
            parent_path: str = this_agent_tool_path.rsplit(".", 1)[0]
            packages = [parent_path]
            resolver = Resolver(packages)

            try:
                python_class = resolver.resolve_class_in_module(class_name, module_name)
            except (ValueError, AttributeError) as second_exception:
                # Get all but the last module in the path.
                # This is what was actually used for AGENT_TOOL_PATH
                agent_tool_path: str = ".".join(this_agent_tool_path.split(".")[:-1])
                agent_network: str = this_agent_tool_path.split(".")[-1]
                agent_name: str = self.factory.get_name_from_spec(self.agent_tool_spec)
                message = f"""
    Could not find class "{class_name}"
    in module "{module_name}"
    under AGENT_TOOL_PATH "{agent_tool_path}"
    for the agent called "{agent_name}"
    in the agent network "{agent_network}".

    Check these things:
    1.  Is there a typo in your AGENT_TOOL_PATH?
    2.  Expected to find a specific CodedTool for the given agent network in:
        <AGENT_TOOL_PATH>/<agent_network>/<coded_tool_name>.py
        Global CodedTools (shared across networks) should be located at:
        <AGENT_TOOL_PATH>/<coded_tool_name>.py
        a)  Does your AGENT_TOOL_PATH point to the correct directory?
        b)  Does your CodedTool actually live in a module appropriately
            named for your agent network?
        c)  Does the module in the "class" designation for the agent {agent_name}
            match what is in the filesystem?
        d)  Does the specified class name match what is actually implemented in the file?
        e)  If an agent network contains both specific and global CodedTools,
            the global module must not have the same name as the agent network.
    """
                raise ValueError(message) from second_exception

        # Instantiate the CodedTool
        coded_tool: CodedTool = None
        try:
            if issubclass(python_class, BranchActivation):
                # Allow for a combination of BranchActivation + CodedTool to allow
                # for easier invocation of agents within code.
                coded_tool = python_class(self.run_context, self.factory,
                                          self.arguments, self.agent_tool_spec, self.sly_data)
            else:
                # Go with the no-args constructor as per the run-of-the-mill contract
                coded_tool = python_class()
        except TypeError as exception:
            message: str = f"""
Coded tool class {python_class} must take no orguments to its constructor.
The standard pattern for CodedTools is to not have a constructor at all.

Some hints:
1)  If you are attempting to re-use/re-purpose your CodedTool implementation,
    consider adding an "args" block to your specific agents. This will pass
    whatever dictionary you specify there as extra key/value pairs to your
    CodedTool's invoke()/async_invoke() method's args parameter in addition
    to those provided by any calling LLM.
2)  If you need something more dynamic that is shared amongst the CodedTools
    of your agent network to handle a single request, consider lazy instantiation
    of the object in question, and share a reference to that  object in the
    sly_data dictionary. The lifetime will of that object will last as long
    as the ruest itself is in motion.
3)  Try very very hard to *not* use global variables/singletons to bypass this limitation.
    Your CodedTool implementation is working in a multi-threaded, asynchronous
    environment. If your first instinct is to reach for a global variable,
    you are highly likely to diminish the performance for all other requests
    on any server running your agent with your CodedTool.
"""
            raise TypeError(message) from exception

        if isinstance(coded_tool, CodedTool):
            # Invoke the CodedTool
            retval: Any = await self.attempt_invoke(coded_tool, self.arguments, self.sly_data)
        else:
            retval = f"Error: {full_class_ref} is not a CodedTool"

        # Change the result into a message
        retval_str: str = f"{retval}"
        message = AIMessage(content=retval_str)

        return message

    async def attempt_invoke(self, coded_tool: CodedTool, arguments: Dict[str, Any], sly_data: Dict[str, Any]) \
            -> Any:
        """
        Attempt to invoke the coded tool.

        :param coded_tool: The CodedTool instance to invoke
        :param arguments: The arguments dictionary to pass as input to the coded_tool
        :param sly_data: The sly_data dictionary to pass as input to the coded_tool
        :return: The result of the coded_tool, whatever that is.
        """
        retval: Any = None

        arguments_dict: Dict[str, Any] = ToolArgumentReporting.prepare_tool_start_dict(arguments)
        message = AgentMessage(content="Received arguments:", structure=arguments_dict)
        await self.journal.write_message(message)

        try:
            # Try the preferred async_invoke()
            retval = await coded_tool.async_invoke(self.arguments, self.sly_data)

        except NotImplementedError:
            # That didn't work, so try running the synchronous method as an async task
            # within the confines of the proper executor.

            # Warn that there is a better alternative.
            message = f"""
Running CodedTool class {coded_tool.__class__.__name__}.invoke() synchronously in an asynchronous environment.
This can lead to performance problems when running within a server. Consider porting to the async_invoke() method.
"""
            self.logger.info(message)
            message = AgentMessage(content=message)
            await self.journal.write_message(message)

            # Try to run in the executor.
            invocation_context = self.run_context.get_invocation_context()
            executor: AsyncioExecutor = invocation_context.get_asyncio_executor()
            loop: AbstractEventLoop = executor.get_event_loop()
            retval = await loop.run_in_executor(None, coded_tool.invoke, arguments, sly_data)

        retval_dict: Dict[str, Any] = {
            "tool_end": True,
            "tool_output": retval
        }
        message = AgentMessage(content="Got result:", structure=retval_dict)
        await self.journal.write_message(message)

        return retval




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from langchain_core.language_models.base import BaseLanguageModel

from neuro_san.internals.run_context.langchain.llms.llm_policy import LlmPolicy


class LangChainLlmResources:
    """
    Class for representing a LangChain model
    together with run-time policy necessary for model usage by the service.
    """

    def __init__(self, model: BaseLanguageModel, llm_policy: LlmPolicy = None):
        """
        Constructor.
        :param model: Language model used.
        :param llm_policy: optional LlmPolicy object to manage connections to LLM host.
        """
        self.model: BaseLanguageModel = model
        self.llm_policy: LlmPolicy = llm_policy

    def get_model(self) -> BaseLanguageModel:
        """
        :return: the BaseLanguageModel
        """
        return self.model

    def get_llm_policy(self) -> LlmPolicy:
        """
        :return: the LlmPolicy used by the model
        """
        return self.llm_policy

    async def delete_resources(self):
        """
        Release the run-time resources used by the model
        """
        if self.llm_policy is not None:
            await self.llm_policy.delete_resources()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.internals.parsers.structure.first_available_structure_parser \
    import FirstAvailableStructureParser
from neuro_san.message_processing.message_processor import MessageProcessor


class StructureMessageProcessor(MessageProcessor):
    """
    Implementation of the MessageProcessor that looks for
    structure information in chat text and extracts structure based on
    the format list that was passed into the constructor.
    """

    def __init__(self, structure_formats: Union[str, List[str]] = None):
        """
        Constructor

        :param structure_formats: Optional string or list of strings telling us to look for
                    specific formats within the text to separate out/extract
                    in a common way so that clients do not have to reinvent this wheel over
                    and over again.

                    Valid values are:
                        "json" - look for JSON in the message content as structure to report.

                    By default this is None, implying that such parsing is bypassed.
        """
        # Only deal with non-empy lists of strings internally
        self.structure_formats: List[str] = structure_formats
        if self.structure_formats is not None:
            if isinstance(self.structure_formats, str):
                self.structure_formats = [self.structure_formats]
        else:
            self.structure_formats = []

        if not isinstance(self.structure_formats, List):
            raise ValueError(f"Value '{structure_formats}' must be a string, a list of strings, or None")

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        text: str = chat_message_dict.get("text")
        structure: Dict[str, Any] = chat_message_dict.get("structure")

        if structure is not None:
            # We already have a structure. Do not modify.
            return

        if text is None or len(text) == 0:
            # There is no text to extract anything from
            return

        # Parse structure from the first available format in the answer content
        structure_parser = FirstAvailableStructureParser(self.structure_formats)
        use_structure: Dict[str, Any] = structure_parser.parse_structure(text)
        if use_structure is None:
            return

        # Modify the existing chat_message_dict to reflect extracted structure
        use_text: str = structure_parser.get_remainder()
        chat_message_dict["text"] = use_text
        chat_message_dict["structure"] = use_structure




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Union

from langchain_core.messages.base import BaseMessage


class AgentFrameworkMessage(BaseMessage):
    """
    BaseMessage implementation of a message from the agent framework
    """

    type: Literal["agent-framework"] = "agent-framework"

    def __init__(self, content: Union[str, List[Union[str, Dict]]] = None,
                 chat_context: Dict[str, Any] = None,
                 sly_data: Dict[str, Any] = None,
                 structure: Dict[str, Any] = None,
                 **kwargs: Any) -> None:
        """
        Pass in content as positional arg.

        :param content: The string contents of the message.
        :param chat_context: A dictionary that fully desbribes the state of play
                    of the chat conversation such that when it is passed on to a
                    different server, the conversation can continue uninterrupted.
        :param sly_data: A dictionary of private data, separate from the chat stream.
        :param structure: A dictionary previously extracted from the content
                        that had been optionally detected by the system as JSON text.
                        The idea is to have the server do the hard parsing so the
                        multitude of clients do not have to rediscover how to best do it.
        :param kwargs: Additional fields to pass to the superclass
        """
        super().__init__(content=content, **kwargs)
        self.chat_context: Dict[str, Any] = chat_context
        self.sly_data: Dict[str, Any] = sly_data
        self.structure: Dict[str, Any] = structure




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.filters.token_accounting_message_filter import TokenAccountingMessageFilter
from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.message_processing.message_processor import MessageProcessor


class TokenAccountingMessageProcessor(MessageProcessor):
    """
    Implementation of the MessageProcessor that looks for the final token accouting
    of the chat session.
    """

    def __init__(self):
        """
        Constructor

        :param structure_formats: Optional string or list of strings telling us to look for
                    specific formats within the text of the final answer to separate out
                    in a common way so that clients do not have to reinvent this wheel over
                    and over again.

                    Valid values are:
                        "json" - look for JSON in the message content as structure to report.

                    By default this is None, implying that such parsing is bypassed.
        """
        self.token_accounting: Dict[str, Any] = None
        self.filter: TokenAccountingMessageFilter = TokenAccountingMessageFilter()

    def get_token_accounting(self) -> Dict[str, Any]:
        """
        :return: The final token accounting from the agent session interaction
        """
        return self.token_accounting

    def reset(self):
        """
        Resets any previously accumulated state
        """
        self.token_accounting = None

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        if not self.filter.allow_message(chat_message_dict, message_type):
            # Does not pass the criteria for a message holding a final answer
            return

        structure: Dict[str, Any] = chat_message_dict.get("structure")

        # Record what we got.
        # We might get another as we go along, but the last message in the stream
        # meeting the criteria above is our final answer.
        if structure is not None:
            self.token_accounting = structure




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

import json
import uuid

from pathlib import Path

from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.internals.messages.origination import Origination
from neuro_san.message_processing.message_processor import MessageProcessor


# pylint: disable=too-many-arguments,too-many-positional-arguments
class ThinkingFileMessageProcessor(MessageProcessor):
    """
    Processes AgentCli input by using the neuro-san streaming API.
    """

    def __init__(self, thinking_file: str, thinking_dir: str):
        """
        Constructor

        :param thinking_file: A string representing the path to a single file where
                              all agent output is combined.  We no longer recommmend this
                              now that we have ...
        :param thinking_dir: A string representing the path to a single directory where
                             each agent in the network gets its own history file according
                             to its agent network origin name.  This is much easier to
                             debug as you do not have to tease apart output from interacting agents.
        """
        self.thinking_file: Path = None
        if thinking_file is not None:
            self.thinking_file = Path(thinking_file)

        self.thinking_dir: Path = None
        if thinking_dir is not None:
            self.thinking_dir = Path(thinking_dir)

        # Dictionary for storing origins that differ from what might be expected
        self.origins: Dict[str, str] = {}

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """

        # Process any text in the message
        text: str = chat_message_dict.get("text")
        structure: Dict[str, Any] = chat_message_dict.get("structure")
        if text is None and structure is None:
            return

        origin: List[str] = chat_message_dict.get("origin")
        test_origin_str: str = Origination.get_full_name_from_origin(origin)

        origin_str: str = ""
        if test_origin_str is not None:
            origin_str = test_origin_str

        self.write_message(chat_message_dict, origin_str)

    def write_message(self, response: Dict[str, Any], origin_str: str):
        """
        Writes a line of text attributable to the origin, however we are doing that.
        :param response: The message to write
        :param origin_str: The string representing the origin of the message
        """

        response_type: str = response.get("type")
        message_type: ChatMessageType = ChatMessageType.from_response_type(response_type)
        message_type_str: str = ChatMessageType.to_string(message_type)

        text: str = response.get("text")
        structure: Dict[str, Any] = response.get("structure")

        if text is None:
            text = ""

        if structure is not None:
            # There is no real text, but there is a structure. JSON-ify it.
            if len(text) > 0:
                text += "\n"
            text += f"```json\n{json.dumps(structure, indent=4, sort_keys=True)}\n```"

        # Figure out how we are going to report the origin given the message.
        use_origin: str = self._determine_origin_reporting(response, origin_str)

        # Determine the filename to use given the origin_str.
        # Previously we might have used a uuid as a filename, but by default
        # we want the filenames to match the full origin_str.
        origin_filename: str = self.origins.get(origin_str, origin_str)

        try:
            self._write_to_file(origin_filename, origin_str, message_type_str, use_origin, text)
        except OSError as os_error:
            # For very deep networks we can sometimes get a "File name too long",
            # which is Error code 63.
            if os_error.errno == 63:

                # Retry with a uuid as file name.
                # If this fails, there's no helping ya.
                origin_filename = str(uuid.uuid4())
                self._write_to_file(origin_filename, origin_str, message_type_str, use_origin, text)

                # Squirell that uuid away so results continue to go to the same
                # file over and over again.
                self.origins[origin_str] = origin_filename
            else:
                raise os_error

    def _write_to_file(self, origin_filename: str, origin_str: str,
                       message_type_str: str, use_origin: str, text: str):

        filename: Path = self.thinking_file
        if self.thinking_dir:
            if origin_filename is None or len(origin_filename) == 0:
                return
            # In case origin_filename is a structured agent name, convert it with "__"
            origin_filename = origin_filename.replace("/", "__")
            filename = Path(self.thinking_dir, origin_filename)

        how_to_open_file: str = "a"
        if not filename.exists():
            how_to_open_file = "w"

        with filename.open(mode=how_to_open_file, encoding="utf-8") as thinking:
            if how_to_open_file == "w":
                # New file, preface it with an origin log.
                thinking.write(f"Agent: {origin_str}\n")

            # Write the message out
            thinking.write(f"\n[{message_type_str}{use_origin}]:\n")
            thinking.write(text)
            thinking.write("\n")

    def _determine_origin_reporting(self, response: Dict[str, Any], origin_str: str) -> str:

        use_origin: str = ""

        # Maybe add some context to where message is coming from if not using thinking_dir
        if not self.thinking_dir:
            use_origin += f" from {origin_str}"

        # Maybe add some context as to where the tool result came from if we have info for that.
        tool_result_origin: List[Dict[str, Any]] = response.get("tool_result_origin")
        if tool_result_origin is not None:
            last_origin_only: List[Dict[str, Any]] = [tool_result_origin[-1]]
            origin_str = Origination.get_full_name_from_origin(last_origin_only)
            use_origin += f" (result from {origin_str})"

        return use_origin
